
# region THetA example
# Generates Sample_T.BEST.results:
# /path/to/theta2/bin/RunTHetA Sample_T.interval_count \
#     --TUMOR_FILE Sample_T.tumor.snp_formatted.txt \
#     --NORMAL_FILE Sample_T.normal.snp_formatted.txt \
#     --BAF --NUM_PROCESSES `nproc` --FORCE
# endregion

# region THetA usage
# usage: RunTHetA.py [-h] [--TUMOR_FILE TUMOR_FILE] [--NORMAL_FILE NORMAL_FILE]
#                    [-n N] [-k K] [-t TAU] [-d DIR] [-p PRE] [-m MAX_NORMAL]
#                    [--NUM_PROCESSES NUM_PROCESSES]
#                    [--NUM_INTERVALS NUM_INTERVALS] [--BOUND_HEURISTIC BH]
#                    [--NORMAL_BOUND_HEURISTIC NBH] [--HEURISTIC_LB LB]
#                    [--HEURISTIC_UB UB] [--BOUNDS_ONLY] [--NO_MULTI_EVENT]
#                    [--RESULTS filename] [--FORCE] [--GET_VALUES]
#                    [--NO_INTERVAL_SELECTION] [--READ_DEPTH_FILE FILENAME]
#                    [--GRAPH_FORMAT GRAPH_FORMAT] [--BAF]
#                    [--RATIO_DEV RATIO_DEV] [--MIN_FRAC MIN_FRAC]
#                    [--NO_CLUSTERING]
#                    QUERY_FILE

# positional arguments:
#   QUERY_FILE            Interval file

# optional arguments:
#   -h, --help            show this help message and exit
#   --TUMOR_FILE TUMOR_FILE
#                         File containing allelic counts for tumor sample SNPs.
#   --NORMAL_FILE NORMAL_FILE
#                         File containing allelic counts for normal samlpe SNPs.
#   -n N, --N N           Number of subpopulations
#   -k K, --MAX_K K       The maximum value allowed for entries in C
#   -t TAU, --TAU TAU     Expected number of copies in normal genome
#   -d DIR, --DIR DIR     Directory where result file is written to
#   -p PRE, --OUTPUT_PREFIX PRE
#                         Prefix for output files created. By default, it will
#                         be the beginning of the input file name (i.e.if input
#                         filename were example.input, output filed would be
#                         example.output and example.withbounds)
#   -m MAX_NORMAL, --MAX_NORMAL MAX_NORMAL
#                         The maximum fraction to consider for normal. Only
#                         enforced for n=2
#   --NUM_PROCESSES NUM_PROCESSES
#                         The number of processes to be used
#   --NUM_INTERVALS NUM_INTERVALS
#                         The maximum number of intervals used by automatic
#                         interval selection.
#   --BOUND_HEURISTIC BH
#   --NORMAL_BOUND_HEURISTIC NBH
#   --HEURISTIC_LB LB
#   --HEURISTIC_UB UB
#   --BOUNDS_ONLY
#   --NO_MULTI_EVENT
#   --RESULTS filename
#   --FORCE
#   --GET_VALUES
#   --NO_INTERVAL_SELECTION
#   --READ_DEPTH_FILE FILENAME
#   --GRAPH_FORMAT GRAPH_FORMAT
#                         Options are .pdf, .jpg, .png, .eps
#   --BAF                 Option to run the BAF model.
#   --RATIO_DEV RATIO_DEV
#                         The deviation away from 1.0 that a ratio must be to be
#                         considered a potential copy number aberration.
#   --MIN_FRAC MIN_FRAC   The minimum fraction of the genome that must have a
#                         potential copy number aberration to be a valid sample
#                         for THetA analysis.
#   --NO_CLUSTERING       Option to run THetA without clustering.
# endregion

# region bam_readcount
# Usage: bam-readcount [OPTIONS] <bam_file> [region]
# Generate metrics for bam_file at single nucleotide positions.
# Example: bam-readcount -f ref.fa some.bam

# Available options:
#   -h [ --help ]                         produce this message
#   -v [ --version ]                      output the version number
#   -q [ --min-mapping-quality ] arg (=0) minimum mapping quality of reads used
#                                         for counting.
#   -b [ --min-base-quality ] arg (=0)    minimum base quality at a position to
#                                         use the read for counting.
#   -d [ --max-count ] arg (=10000000)    max depth to avoid excessive memory
#                                         usage.
#   -l [ --site-list ] arg                file containing a list of regions to
#                                         report readcounts within.
#   -f [ --reference-fasta ] arg          reference sequence in the fasta format.
#   -D [ --print-individual-mapq ] arg    report the mapping qualities as a comma
#                                         separated list.
#   -p [ --per-library ]                  report results by library.
#   -w [ --max-warnings ] arg             maximum number of warnings of each type
#                                         to emit. -1 gives an unlimited number.
#   -i [ --insertion-centric ]            generate indel centric readcounts.
#                                         Reads containing insertions will not be
#                                         included in per-base counts
# endregion

# region snp formted file
# (1) Chrm
# (2) pos
# (3) A -  # reads with A at this position
# (4) C - # reads with C at this position
# (5) G - # reads with G at this position
# (6) T - # reads with T at this position
# (7) Total - the # reads at this position
# (8) refCount - # reads with the reference allele (indicated in SNP file)
# (9) mutCount - # reads with the mutant allele (indicated in SNP file)
# endregion

from pyppl import Box
from os import path, makedirs
from bioprocs.utils import logger, shell2 as shell
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter, TsvRecord
from bioprocs.utils.parallel import Parallel, distribute
from bioprocs.utils.reference import bamIndex
from bioprocs.utils.shell2 import ln_s, wc_l

{% python from os import path %}
itvfile  = {{ i.infile | quote}}
prefix   = {{ o.outdir | path.join: fn(i.infile) | quote }}
tumbam   = {{ i.tumbam | quote}}
tumsnp   = {{ o.outdir | path.join: fn(i.tumbam) | @append: '.bamrc' | quote}}
normbam  = {{ i.normbam | quote}}
normsnp  = {{ o.outdir | path.join: fn(i.normbam) | @append: '.bamrc' | quote}}
outfile  = {{ o.outfile | quote}}
outdir   = {{ o.outdir | quote}}
params   = {{ args.params | repr}}
ref      = {{ args.ref | quote}}
bedtools = {{ args.bedtools | quote }}
bamrc    = {{ args.bam_readcount | quote }}
nthread  = {{ args.nthread | repr}}
affysnps = {{ args.affysnps | quote}}
theta    = {{ args.theta | quote}}
samtools = {{ args.samtools | quote}}

shell.load_config(
	theta2        = Box(_exe = theta, _raw = True),
	samtools      = samtools,
	bedtools      = bedtools,
	bam_readcount = bamrc,
)
if tumbam.endswith('.bam'):
	logger.info('- Try to index tumor bam file...')
	bamIndex(tumbam,  samtools = samtools, nthread = nthread)
if normbam.endswith('.bam'):
	logger.info('- Try to index normal bam file...')
	bamIndex(normbam, samtools = samtools, nthread = nthread)

if not path.isfile(affysnps) and (tumbam.endswith('.bam') or normbam.endswith('.bam')):
	raise ValueError('Formatted SNPs for tumor or normal not provided, so snp list is required.')

if itvfile.endswith('.bed') and not (tumbam.endswith('.bam') and normbam.endswith('.bam')):
	raise ValueError('Bam files of tumor and normal required as interval file is a BED.')

def getCoverage(bamfiles, region, outfile):
	"""
	Get coverage of regions
	"""
	logger.info('- Get read counts from %s ...' % bamfiles)
	if not region.endswith('.bed'): # read counts already provided
		logger.info('  Read counts already provided in region file, skip.')
		ln_s(bamfile, outfile)
		return

	writer = TsvWriter(outfile)
	writer.cnames = ['ID', 'Chrm', 'Start', 'End', 'numTumor', 'numNormal']
	writer.writeHead(lambda ns: '#' + '\t'.join(ns))
	i = 0
	for line in shell.bedtools.multicov(
		bams  = bamfiles,
		bed   = region,
		_iter = True):
		# chr1	1	12444	104	87
		i += 1
		parts = line.strip().split('\t')
		parts.insert(0, i)
		writer.write(parts[:4] + parts[-2:])
	writer.close()

def getAlleleCount(bamfile, snpfile, outfile):
	logger.info('- Get allele counts from %s ...' % bamfile)
	if not bamfile.endswith('.bam'): # it's THetA2 formatted SNP file already
		logger.info('  THetA2 formatted SNP file already, skip.')
		ln_s(bamfile, outfile)
		return
	brcparams       = Box()
	brcparams.f     = ref
	brcparams.w     = 0
	brcparams.l     = snpfile
	brcparams._iter = True

	brcparams[''] = bamfile

	writer = TsvWriter(outfile)
	writer.cnames = ['Chrm', 'pos', 'A', 'C', 'G', 'T', 'Total', 'refCount', 'mutCount']
	writer.writeHead(lambda ns: '#' + '\t'.join(ns))
	for line in shell.bam_readcount(**brcparams):
		#chr1	564773	C	14	=:0:0.00:0.00:0.00:0:0:0.00:0.00:0.00:0:0.00:0.00:0.00	A:0:0.00:0.00:0.00:0:0:0.00:0.00:0.00:0:0.00:0.00:0.00	C:14:...	G:0:...	T:0:...	N:0:...
		r = line.split('\t')
		counts = dict(
			A = r[5].split(':', 2)[1],
			C = r[6].split(':', 2)[1],
			G = r[7].split(':', 2)[1],
			T = r[8].split(':', 2)[1]
		)
		rec    = TsvRecord()
		rec.Chrm  = r[0]
		rec.pos   = r[1]
		rec.Total = r[3]
		rec.A = counts['A']
		rec.C = counts['C']
		rec.G = counts['G']
		rec.T = counts['T']
		# if reference allele is unknown, assuming all are mut alleles
		refallele = r[2].upper()
		rec.refCount = counts.get(refallele, 0)
		counts = [int(count) for allele, count in counts.items() if allele != refallele and count.isdigit()]
		rec.mutCount = max(counts) if counts else 0
		writer.write(rec)
	writer.close()

if nthread == 1:
	getAlleleCount(tumbam, affysnps, tumsnp)
	getAlleleCount(normbam, affysnps, normsnp)
else:
	# try to split the affysnps into N files and distribute the jobs to nthreads
	# get number of lines of affysnps file
	total  = int(wc_l(affysnps).split()[0])
	dists  = distribute(total, nthread)
	reader = TsvReader(affysnps, cnames = False)
	# dir to save the split file and result file
	thdir  = path.join(outdir, 'bamrc.nthreads')
	if not path.exists(thdir):
		makedirs(thdir)

	asbname = path.basename(affysnps).split('.')[0]
	for i, dist in enumerate(dists):
		writer = TsvWriter(path.join(thdir, '{bname}.thread{i}.snp'.format(
			bname = asbname, i = i
		)))
		for _ in range(dist):
			writer.write(next(reader))
		writer.close()

	para   = Parallel(nthread, raiseExc = True)
	para.run(getAlleleCount, [
		(tumbam, path.join(
			thdir, '{bname}.thread{i}.snp'.format(bname = asbname, i = i)
		), path.join(
			thdir, '{tumbn}.thread{i}.bamrc'.format(tumbn = path.basename(tumbam), i = i)
		)) for i in range(nthread)
	])
	# merge to tumsnp
	writer = TsvWriter(tumsnp)
	writer.cnames = ['Chrm', 'pos', 'A', 'C', 'G', 'T', 'Total', 'refCount', 'mutCount']
	writer.writeHead(lambda cn: "#" + "\t".join(cn))
	for i in range(nthread):
		subrc = path.join(
			thdir, '{tumbn}.thread{i}.bamrc'.format(tumbn = path.basename(tumbam), i = i)
		)
		reader = TsvReader(subrc, cnames = False)
		for r in reader:
			writer.write(r.values())
		reader.close()
	writer.close()

	# normal
	para.run(getAlleleCount, [
		(normbam, path.join(
			thdir, '{bname}.thread{i}.snp'.format(bname = asbname, i = i)
		), path.join(
			thdir, '{normbn}.thread{i}.bamrc'.format(normbn = path.basename(normbam), i = i)
		)) for i in range(nthread)
	])
	# merge to normsnp
	writer = TsvWriter(normsnp)
	writer.cnames = ['Chrm', 'pos', 'A', 'C', 'G', 'T', 'Total', 'refCount', 'mutCount']
	writer.writeHead(lambda cn: "#" + "\t".join(cn))
	for i in range(nthread):
		subrc = path.join(
			thdir, '{normbn}.thread{i}.bamrc'.format(normbn = path.basename(normbam), i = i)
		)
		reader = TsvReader(subrc, cnames = False)
		for r in reader:
			writer.write(r.values())
		reader.close()
	writer.close()

# check type of itvfile
if itvfile.endswith('.bed'):
	tmpfile = path.join(outdir, path.basename(itvfile).split('.')[0] + '.interval.txt')
	getCoverage([tumbam, normbam], itvfile, tmpfile)
	itvfile = tmpfile


params.NUM_PROCESSES = nthread
params.TUMOR_FILE    = tumsnp
params.NORMAL_FILE   = normsnp
params.GRAPH_FORMAT  = '.png'
params.OUTPUT_PREFIX = prefix
#params.RATIO_DEV     = params.get('RATIO_DEV', .5)
#params.MIN_FRAC      = params.get('MIN_FRAC', .1)
params['']           = itvfile
shell.fg.theta2(**params)

ln_s(prefix + '.BEST.results', outfile)





import logging
import sys
import shlex
import re
import gzip
from glob import glob
from os import path
import yaml
from pyppl.utils import Box

class Mem2Exception(Exception):
	pass

def ensureBox(var):
	if isinstance(var, Box):
		return var
	return Box(var)

def fs2name(files):
	if not files: return 'nothing.etc'
	bnames  = [path.basename(f).split('.')[0] for f in files]
	compfix = path.commonprefix(bnames)
	compfix = re.sub(r'[^a-zA-Z0-9]$', '', compfix)
	if len(compfix) < 3:
		return bnames[0] + '.etc'
	else:
		return compfix + '.etc'

def reportdata(outdir, *args, **kwargs):
	data = {}
	for arg in args:
		if not isinstance(arg, dict):
			raise TypeError('Expect a dict as report data.')
		data.update(arg)
	data.update(kwargs)
	datafile = path.join(outdir, 'job.report.data.yaml')
	with open(datafile, 'w') as f:
		yaml.dump(data, f, default_flow_style=False)

def dirpat2name(directory, pattern = '*'):
	return fs2name(glob(path.join(directory, pattern)))

def getLogger(name = 'bioprocs', logfmt = "[%(asctime)s %(levelname)7s] %(message)s", datafmt = "%Y-%m-%d %H:%M:%S", level = logging.INFO):
	logger = logging.getLogger(name)
	for handler in logger.handlers:
		handler.close()
	del logger.handlers[:]

	ch = logging.StreamHandler()
	ch.setFormatter(logging.Formatter(logfmt, datafmt))
	logger.addHandler(ch)

	logger.setLevel(level)
	return logger

def log2pyppl(msg, level = None):
	level = '.' + level if level else ''
	sys.stderr.write('pyppl.log%s:%s\n' % (level, msg))

def _autoUnit(num):
	if num % (1024 * 1024) == 0:
		return num / (1024*1024), 'G'
	elif num % 1024 == 0:
		return num / 1024, 'M'
	else:
		return num, 'K'

# unit = 'auto'/'G'/'M'/'K'/'java'
def mem2 (mem, unit = 'auto'):
	mem   = str(mem)
	ounit = mem[-1].upper()
	if ounit == 'G':
		num   = int(mem[:-1]) * 1024 * 1024
	elif ounit == 'M':
		num   = int(mem[:-1]) * 1024
	elif ounit == 'K':
		num   = int(mem[:-1])
	elif not ounit.isdigit():
		raise Mem2Exception('Unknown memory unit: ' + ounit)
	else:
		num   = int(mem)

	unit = unit.upper()
	retn, retu = _autoUnit(num)
	if unit == 'G':
		if retu == 'M':   retn /= 1024.
		elif retu == 'K': retn /= (1024. * 1024.)
		retu = 'G'
	elif unit == 'M':
		if retu == 'G':   retn *= 1024
		elif retu == 'K': ret /= 1024.
		retu = 'M'
	elif unit == 'K':
		if retu == 'G':   retn *= 1024 * 1024
		elif retu == 'M': retn *= 1024
		retu = 'K'

	if unit == 'JAVA':
		xmx = "-Xmx" + str(int(retn)) + retu
		n, u = _autoUnit(num / 8)
		return '-Xms' + str(int(n)) + u + ' ' + xmx
	elif unit == 'JAVADICT' or unit == 'JDICT':
		n, u = _autoUnit(num / 8)
		return {
			'Xmx' + str(retn) + retu: True,
			'Xms' + str(n) + u: True
		}
	elif unit == '-JAVADICT' or unit == '-JDICT':
		n, u = _autoUnit(num / 8)
		return {
			'-Xmx' + str(retn) + retu: True,
			'-Xms' + str(n) + u: True
		}
	else:
		return str(retn) + retu

def regionOverlap(CHR1, START1, END1, CHR2, START2, END2):
	if CHR1 != CHR2: return False
	if int(END1) < int(START2): return False
	if int(END2) < int(START1): return False
	return True

def funcargs(func):
	if not callable(func):
		raise ValueError('Expect a callable.')
	try:
		from inspect import signature
		return list(signature(func).parameters.keys())
	except ImportError:
		from inspect import getargspec
		return getargspec(func).args

def gztype(gzfile):
	import binascii
	with open(gzfile, 'rb') as f:
		flag = binascii.hexlify(f.read(4))
	if flag == b'1f8b0804':
		return 'bgzip'
	if flag == b'1f8b0808':
		return 'gzip'
	return 'flat'

class FileConn(object):

	def __init__(self, filename, flag = 'r'):
		self.isgz = False
		if filename.endswith('.gz'):
			self.isgz = True
		self.filename = filename
		self.flag     = flag
		self.fp       = None

	def __enter__(self):
		return self.open()

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.close()

	def open(self):
		if self.isgz:
			self.fp = gzip.open(self.filename, self.flag)
		else:
			self.fp = open(self.filename, self.flag)
		return self.fp

	def close(self):
		self.fp.close()

def switch(value, cases, *args, **kwargs):
	ret = cases.get(value, 'default')
	if callable(ret) and (args or kwargs):
		return ret(*args, **kwargs)
	return ret

logger = getLogger()



from sys import stderr
from pyppl import Box
from os import path
from bioprocs.utils import shell, mem2
from bioprocs.utils.poll import Poll
from bioprocs.utils.shell import Shell
from bioprocs.utils.reference import bamIndex

infile     = {{ i.infile | quote}}
bamIndex(infile, samtools = None)
outfile    = {{ o.outfile | quote}}
outprefix  = {{ o.outfile | prefix | quote}}
tool       = {{ args.tool | quote}}
gatk       = {{ args.gatk | quote}}
samtools   = {{ args.samtools | quote}}
picard     = {{ args.picard | quote}}
bamutil    = {{ args.bamutil | quote}}
params     = {{ args.params | repr}}
ref        = {{ args.ref | quote}}
tmpdir     = {{ args.tmpdir | quote}}
knownSites = {{ args.knownSites | quote}}
argsmem    = {{ args.mem | quote}}
nthread    = {{ args.nthread | quote}}
joboutdir  = {{ job.outdir | quote}}
joblen     = {{ proc.size | repr}}
jobindex   = {{ job.index | repr}}
workdir    = {{ proc.workdir | quote}}
tmpdir     = path.join (tmpdir, "{{proc.id}}.{{i.infile | fn}}.{{job.index}}")
# pass the index file
shell.ln_s(infile + '.bai', outfile + '.bai')
if not path.exists (tmpdir):
	shell.mkdir (tmpdir, p = True)
shell.TOOLS.update(dict(
	gatk     =  gatk,
	samtools = samtools,
	bamutil  = bamutil
))

# checks
if tool == 'gatk' and not path.isfile(knownSites):
	raise ValueError("knownSites file is required by GATK but is not specified (args.knownSites) or not exists.")

def run_gatk():
	mem = mem2(argsmem, '-jdict')
	intfile = path.join(joboutdir, outprefix + '.intervals')

	mem['-Djava.io.tmpdir={}'.format(shell.shquote(tmpdir))] = True
	gatksh = Shell(equal = ' ', dash = '-').gatk

	rtcparams    = params.get('RealignerTargetCreator', Box())
	rtcparams.T  = 'RealignerTargetCreator'
	rtcparams.R  = ref
	rtcparams.I  = infile
	rtcparams.o  = intfile
	rtcparams.nt = nthread
	rtcparams._  = list(mem.keys())
	gatksh(**rtcparams).run()

	bamfileir    = path.join(joboutdir, outprefix + '.ir.bam')
	irparams     = params.get('IndelRealigner', Box())
	irparams.T   = 'IndelRealigner'
	irparams.R   = ref
	irparams.I   = infile
	irparams.o   = bamfileir
	irparams._   = list(mem.keys())

	irparams.targetIntervals = intfile
	gatksh(**irparams).run()

	recaltable   = path.join(joboutdir, outprefix + '.recaltable')
	brparams     = params.get('BaseRecalibrator', Box())
	brparams.T   = 'BaseRecalibrator'
	brparams.R   = ref
	brparams.I   = bamfileir
	brparams.o   = recaltable
	brparams.nct = nthread
	brparams._   = list(mem.keys())

	brparams.knownSites = knownSites
	gatksh(**brparams).run()

	prparams     = params.get('PrintReads', Box())
	prparams.T   = 'PrintReads'
	prparams.R   = ref
	prparams.I   = bamfileir
	prparams.o   = outfile
	prparams.nct = nthread
	prparams._   = list(mem.keys())
	gatksh(**prparams).run()

	shell.rm(bamfileir, f = True)
	shell.mv(outprefix + '.bai', outfile + '.bai')

def run_bamutil():
	bush = Shell(subcmd = True, equal = ' ').bamutil
	if knownSites:
		params.dbsnp = knownSites
	params['in'] = infile

	params.out     = outfile
	params.refFile = ref
	params.verbose = True
	refcache       = path.splitext(ref)[0] + '-bs.umfa'
	if not path.isfile(refcache):
		poll = Poll(workdir, joblen, jobindex)
		poll.first(lambda **kwargs: bush.recal(**kwargs).run(), **params)
	else:
		bush.recab(**params).run()

tools = dict(
	gatk    = run_gatk,
	bamutil = run_bamutil
)
try:
	tools[tool]()
except KeyError:
	raise KeyError('Tool {!r} not supported.'.format(tool))
except Exception as ex:
	stderr.write ("Job failed: %s" % str(ex))
	raise
finally:
	shell.rm(tmpdir, r = True, f = True)



samples = []
ngenes  = 0
with open("{{i.expfile}}") as f:
	for line in f:
		if not samples:
			samples = line.strip().split('\t')
		elif line.strip():
			ngenes += 1

with open("{{i.expfile}}") as f, open("{{o.outfile}}", "w") as fout:
	fout.write("#1.2\n")
	fout.write("%s\t%s\n" % (ngenes, len(samples)))
	fout.write("NAME\tDescription\t%s\n" % ('\t'.join(samples)))
	f.readline() # header
	for line in f:
		line  = line.strip()
		if not line: continue
		parts = line.split('\t')
		parts.insert(1, parts[0])
		fout.write('\t'.join(parts) + '\n')



from os import path
from copy import deepcopy
from pyppl import Box
from bioprocs.utils import shell2 as shell

cnvkit   = {{args.cnvkit | quote}}
cnrfile  = {{i.cnrfile | quote}}
cnsfile  = {{i.cnsfile | quote}}
outdir   = {{o.outdir | quote}}
regions  = {{args.regions | list}}
nthread  = {{args.nthread | repr}}
params   = {{args.params | repr}}

shell.load_config(cnvkit = dict(
	_exe = cnvkit,
	_env = dict(
		OPENBLAS_NUM_THREADS = str(nthread),
		OMP_NUM_THREADS      = str(nthread),
		NUMEXPR_NUM_THREADS  = str(nthread),
		MKL_NUM_THREADS      = str(nthread)
	),
	_cwd = outdir
))

for region in regions:
	if not region:
		rgion = ''
		genes = ''
	else:
		parts = region.split(':')
		if len(parts) == 1:
			rgion = parts[0]
		elif len(parts) == 2:
			import re
			if re.match(r'^\d+\-\d+$', parts[1]):
				rgion = ':'.join(parts)
				genes = ''
			else:
				rgion = parts[0]
				genes = parts[1]
		else:
			rgion = ':'.join(parts[:2])
			genes = parts[2]

	stem = region.replace(':', '_').replace(',', '_')
	if not stem: stem = 'whole-genome'

	outfile   = path.join(outdir, stem + '.pdf')
	iparams   = deepcopy(params)
	iparams.o = outfile

	if cnsfile:
		iparams.s = cnsfile

	if rgion:
		iparams.c = rgion

	if genes:
		iparams.g = genes

	shell.fg.cnvkit.scatter(cnrfile, **iparams)



import binascii, sys
from pyppl import Box
from bioprocs.utils import runcmd, cmdargs

def gzip_type(fn):
	with open(fn, 'rb') as f:
		flag = binascii.hexlify(f.read(4))
	if flag == b'1f8b0804':
		return 'bgzip'
	if flag == b'1f8b0808':
		return 'gzip'
	return 'flat'

def gunzip(fn, outfn = None):
	args = {'f': True, '1': True}
	if not outfn:
		cmd = 'gunzip {!r} {}'.format(fn, args)
	else:
		cmd = 'gunzip {!r} {} -c > {!r}'.format(fn, args, outfn)
	runcmd(cmd)

def bgzip(fn, outfn = None):
	if not outfn:
		cmd = 'bgzip {!r}'.format(fn)
	else:
		cmd = 'bgzip {!r} -c > {!r}'.format(fn, outfn)
	runcmd(cmd)

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}

gz     = {{args.gz | repr}}
params = {{args.params | repr}}
tabix  = {{args.tabix | quote}}

# make sure outfile without .gz
if gz: outfile = outfile[:-3]

gztype = gzip_type(infile)

if gztype == 'gzip':
	gunzip(infile, outfile)
	if gz:
		bgzip(outfile)
elif gztype == 'bgzip':
	if gz:
		sys.symlink(infile, outfile + '.gz')
	else:
		gunzip(infile, outfile)
else:
	if gz:
		bgzip(infile, outfile + '.gz')
	else:
		sys.symlink(infile, outfile)

cmd = '{} {} {!r}'.format(tabix, cmdargs({'p': 'vcf'}), outfile + '.gz' if gz else outfile)
runcmd(cmd)



from bioprocs.utils.meme import MemeReader, MemeWriter

infile  = {{ i.infile | quote}}
outfile = {{ o.outfile | quote}}
# if filter has multiple lines, treat the first lines as helper and last line as the funciton
# now load the first lines
{{ args.filter | :"" if not "\n" in a else "\n".join(a.split("\n")[:-1]) }}
ffunc   = {{ args.filter | :None if a is None else a.split("\n")[-1] }}

reader = MemeReader(infile)
writer = MemeWriter(outfile)
writer.meta = reader.meta
writer.writeMeta()

for r in reader:
	if callable(ffunc) and ffunc(r):
		continue
	writer.write(r)
reader.close()
writer.close()



from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell

cnvkit   = {{args.cnvkit | quote}}
infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
params   = {{args.params | repr}}
nthread  = 1

shell.load_config(cnvkit = dict(
	_exe = cnvkit,
	_env = dict(
		OPENBLAS_NUM_THREADS = str(nthread),
		OMP_NUM_THREADS      = str(nthread),
		NUMEXPR_NUM_THREADS  = str(nthread),
		MKL_NUM_THREADS      = str(nthread)
	),
	_cwd = path.dirname(outfile)
))

params.o = outfile
shell.fg.cnvkit.call(infile, **params)



"""Operations on TCGA MAF file"""
from pyppl import Proc, Box
from . import params, rimport
from .utils import fs2name
# Next-step processing after VCF file being generated.
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

from .tumhet import _pTMBurden

@procfactory
def _pGTMatAddRs():
	"""
	@name:
		pGTMatAddRs
	@description:
		Add rs id to a genotype matrix
	@input:
		infile: The input genotype matrix, columns are samples, rows are mutations in format:
			- `<chr>_<pos>_<ref>_<alt>` or `<chr>_<pos>_<name>_<ref>_<alt>`
			- has to be sorted by coordinates.
			- `chromsomes` have to be in order of `args.chrorder`
	@output:
		outfile: The output genotype matrix. Row names will turn into:
			- `<chr>_<pos>_<rs>_<ref>_<alt>`
	@args:
		`dbsnp`: the dbsnp vcf file used to annotation the snps.
			- assume sorted by coordinates
			- `chromsomes` have to be in order of `args.chrorder`
		`notfound`: What to used if RS id not found. Default: `NOVEL`
			- `None/Fase` to skip to record
		`exist`: What if RS id exists? Default: `keep`
			- `keep`: Keep the RS ID and skip seeking
			- `force`: Force using the RS ID being found to replace the old one.
		`chrorder`: The chromsome order. Default: `<params.chrorder>`
	"""
	pGTMatAddRs               = Proc(desc = "Add rs id to a genotype matrix")
	pGTMatAddRs.input         = 'infile:file'
	pGTMatAddRs.output        = 'outfile:file:{{i.infile | bn}}'
	pGTMatAddRs.args.dbsnp    = params.dbsnp_all.value
	pGTMatAddRs.args.chrorder = params.chrorder.value
	pGTMatAddRs.args.notfound = 'NOVEL'
	pGTMatAddRs.args.exist    = 'keep'
	pGTMatAddRs.lang          = params.python.value
	pGTMatAddRs.script        = "file:scripts/tcgamaf/pGTMatAddRs.py"
	return pGTMatAddRs

@procfactory
def _pGTMat2Plink():
	"""
	@name:
		pGTMat2Plink
	@description:
		Convert a genotype matrix to plink binary files
	@input:
		infile: The genotype matrix, probably generated by `pVcf2GTMat`
		`metafile:file`: The metadata file.
			- column names could be `['FID', 'IID', 'PID', 'MID', 'Sex', 'Pheno']`, see plink's `ped` format
			- row names are samples
	@output:
		`outdir:dir`: The output directory. Default: `{{i.infile | fn}}.plink`
	@args:
		`plink`:   The path to `plink`
		`keeptxt`: Keep the text files (.ped and .map) or not. Default: `False`
	@requires:
		`plink 1.x`
	"""
	pGTMat2Plink              = Proc(desc = 'Convert a genotype matrix to plink binary files')
	pGTMat2Plink.input        = 'infile:file, metafile:file'
	pGTMat2Plink.output       = 'outdir:dir:{{i.infile | fn}}.plink'
	pGTMat2Plink.args.plink   = params.plink.value
	pGTMat2Plink.args.keeptxt = False
	pGTMat2Plink.args.chrmaps = {'X': 23, 'Y': 24, 'XY': 25, 'M': 26, 'MT': 26}
	pGTMat2Plink.lang         = params.python.value
	pGTMat2Plink.script       = "file:scripts/tcgamaf/pGTMat2Plink.py"
	return pGTMat2Plink

@procfactory
def _pGTMat2Bed():
	"""
	@name:
		pGTMat2Bed
	@description:
		Convert a genotype matrix to a bed file containing the coordinates of the mutations
	@input:
		infile: The genotype matrix. Row names must follow `<chr>_<pos>_<rsid>_<ref>_<alt>`
	@output:
		outfile: The output bed file. Default: `outfile:file:{{i.infile | fn}}.bed`
	@args:
		`ncol`  : How many columns of bed to output. Default: `6`.
			- Possible values: 3, 6, 8, 66, 88
			- If `8`, then reference and alternative alleles will be 7th and 8th column
			- If `66`, then genotypes will be attached to BED6
			- If `88`, then genotypes will be attached to `ncol = 8`
		`name`  : Use the neat name (usually rsid) or full name (row names). Default: `neat`
		`inopts`: Options to read the input file. Default: `Box(cnames = True)`
	"""
	pGTMat2Bed             = Proc(desc = 'Convert a genotype matrix to bed file')
	pGTMat2Bed.input       = 'infile:file'
	pGTMat2Bed.output      = 'outfile:file:{{i.infile | fn}}.bed'
	pGTMat2Bed.args.inopts = Box(cnames = True)
	pGTMat2Bed.args.ncol   = 6
	pGTMat2Bed.args.name   = 'neat' # full
	pGTMat2Bed.lang        = params.python.value
	pGTMat2Bed.script      = "file:scripts/tcgamaf/pGTMat2Bed.py"
	return pGTMat2Bed

@procfactory
def _pCallRate():
	"""
	@name:
		pCallRate
	@description:
		Calculate sample/snp call rate from single sample vcfs
	@input:
		indir:     The dir containing the vcfs
	@output:
		`outsample:file`: The report of call rate for each sample
		`figsample:file`: The bar chat of sample call rates
		`outsnp:file`:    The report of call rate for each snp
		`figsnp:file`:    The bar chat of snp call rates
	"""
	pCallRate                  = Proc()
	pCallRate.input            = "indir:file"
	pCallRate.output           = "outdir:dir:{{ i.indir | fn }}.callrate"
	pCallRate.args.histplotggs = []
	pCallRate.args.devpars     = Box({'res':300, 'width':2000, 'height':2000})
	pCallRate.lang             = params.Rscript.value
	pCallRate.script           = "file:scripts/tcgamaf/pCallRate.r"
	return pCallRate

@procfactory
def _pCepip():
	"""
	@name:
		pCepip
	@description:
		Run CEPIP.
	@input:
		infile: The input file (vcf or avinput)
	@output:
		outfile: The cepip result file
	@args:
		`cepip`:    The path of cepip
		`cell` :    The related cell line
		`params`:   Other params for cepip
	@requires:
		[`cepip`](http://jjwanglab.org/cepip/)
	"""
	pCepip                     = Proc(desc = 'Run cepip for input mutations.')
	pCepip.input               = "infile:file"
	pCepip.output              = "outfile:file:{{i.infile | fn}}.cepip.txt"
	pCepip.args.cepip          = params.cepip.value
	pCepip.args.cell           = ""
	pCepip.args.params         = Box()
	pCepip.lang                = params.python.value
	pCepip.script              = "file:scripts/tcgamaf/pCepip.py"
	return pCepip

@procfactory
def _pMutSig():
	"""
	@name:
		pMutSig
	@description:
		MutSig stands for "Mutation Significance".  MutSig analyzes lists of mutations discovered in DNA sequencing, to identify genes that were mutated more often than expected by chance given background mutation processes.
		For more information, see Lawrence, M. et al. Mutational heterogeneity in cancer and the search for new cancer-associated genes. Nature 499, 214-218 (2013).
		See [dcumentation](http://archive.broadinstitute.org/cancer/cga/mutsig_run)
	@input:
		infile: mutation table
	@output:
		`outdir:dir`: The output directory
	@args:
		`mutsig` : The path to `run_MutSigCV.sh`, default: 'mutsig'
		`mcr`    : The Matlab MCR path
		`cvrg`   : coverage table
		`cvrt`   : covariates table
		`mutdict`: mutation_type_dictionary_file
		`chrdir` : chr_files_hg18 or chr_files_hg19
	@requires:
		[MutSig](http://archive.broadinstitute.org/cancer/cga/mutsig_download)
	"""
	pMutSig              = Proc(desc = 'Run MutSig.')
	pMutSig.input        = 'infile:file'
	pMutSig.output       = "outdir:dir:{{i.infile | fn}}.mutsig"
	pMutSig.args.cvrg    = params.mutsig_cvrg.value
	pMutSig.args.cvrt    = params.mutsig_cvrt.value
	pMutSig.args.mutdict = params.mutsig_mutdict.value
	pMutSig.args.chrdir  = params.mutsig_chrdir.value
	pMutSig.args.mutsig  = params.mutsig.value
	pMutSig.args.mcr     = params.mcr.value
	pMutSig.script       = "file:scripts/tcgamaf/pMutSig.bash"
	return pMutSig

@procfactory
def _pMafLiftover():
	"""
	@name:
		pMafLiftover
	@description:
		Liftover maf file from one assembly to another
	@input:
		infile: The input maf file
	@output:
		outfile: The output maf file
	@args:
		`liftover`: The liftOver program.
		`lochain`:  The liftOver chain file.
		`genome`:   The target genome.
	@requires:
		liftOver from UCSC
	"""
	pMafLiftover               = Proc(desc = 'Liftover a maf file from one assembly to another')
	pMafLiftover.input         = 'infile:file'
	pMafLiftover.output        = 'outfile:file:{{i.infile | fn | lambda x: x if x.endswith(".maf") else x + ".maf"}}'
	pMafLiftover.args.liftover = params.liftover.value
	pMafLiftover.args.lochain  = params.lochain.value
	pMafLiftover.args.genome   = params.genome.value
	pMafLiftover.lang          = params.python.value
	pMafLiftover.script        = "file:scripts/tcgamaf/pMafLiftOver.py"
	return pMafLiftover


@procfactory
def _pMafMerge():
	"""
	@input:
		`infiles:files`: The maf files
	@output:
		outfile: The merged maf file
	@args:
		`excols`: How to deal with extra columns other than 34 standard columns from TCGA.
			- merge(default): Merge the columns, if one not exists, fill with __UNKNOWN__.
			- discard: Just discard the extra columns, with only 34 columns left. So you can also put just one maf file in the indir with some columns missed to fill it with standard columns.
	"""
	return Box(
		desc   = 'Merge MAF files',
		input  = 'infiles:files',
		output = 'outfile:file:{{i.infiles | fs2name}}.maf',
		lang   = params.python.value,
		args   = Box(excols  = 'merge'),  # discard,
		envs   = Box(fs2name = fs2name)
	)

@procfactory
def _pMaf2Mat():
	"""
	@name:
		pMaf2Mat
	@description:
		Convert maf file to a gene(row)-sample(column) matrix
	@input:
		infile: The input file
	@output:
		outfile: The output matrix
	@args:
		`mutypes`: Provide manual list of variant classifications to be counted, only effective when `args.binary = False`. Default: `None` (all counted)
		`binary` : Just generate a binary matrix instead of a count matrix. Default: `False`
		`na`: What value to use for no mutations reported on a gene. Default: `0`
		`samfn`  : A function (in r) to transform the sample names. Default: `function(sample) sample`
	"""
	pMaf2Mat              = Proc(desc = 'Convert maf file to a gene-based mutation matrix')
	pMaf2Mat.input        = 'infile:file'
	pMaf2Mat.output       = 'outfile:file:{{i.infile | fn}}.mat.txt'
	pMaf2Mat.args.binary  = False
	pMaf2Mat.args.mutypes = None
	pMaf2Mat.args.na      = 0
	pMaf2Mat.args.samfn   = 'function(sample) sample'
	pMaf2Mat.lang         = params.Rscript.value
	pMaf2Mat.script       = "file:scripts/tcgamaf/pMaf2Mat.r"
	return pMaf2Mat

@procfactory
def _pMaftools():
	"""
	@input:
		indir: The input directory or a single maf file. A directory could contain:
			- `*.maf` or `*.maf.gz` file (required)
			- `*.annot.tsv` or `*.annot.txt` file (see: https://github.com/PoisonAlien/maftools/blob/master/inst/extdata/tcga_laml_annot.tsv)
			- `all_lesions.conf_*.txt`: Gistic cnv data
			- `amp_genes.conf_*.txt`: Gistic cnv data
			- `del_genes.conf_*.txt`: Gistic cnv data
			- `scores.gistic`: Gistic cnv data
			- `*.seg.txt`: CBS segments data
		msdir: MutSig result directory.
			- Must have`*sig_genes.txt` to do pancancer comparison.
	@output:
		outdir: The output directory
	@args:
		ngenes : Top number of genes to plot for some plots.
		extypes: Exclude mutation types, only consider nonsynonymous mutations.
		isTCGA : If the maf file is from TCGA?
		ref    : The reference file for signature plot.
		plot   : Which plots to plot.
		params : The extra parameters for each plot function.
		devpars: The parameters for plot device.
		nthread: Number of threads used for multiple plot of one type.
	@requires:
		[Maftools](https://bioconductor.org/packages/devel/bioc/vignettes/maftools/inst/doc/maftools.html)
	"""
	return Box(
		desc   = 'Basic analysis on somatic mutations of a group of samples',
		input  = 'indir:file, msdir:file',
		output = 'outdir:dir:{{i.indir | fn}}.maftools',
		lang   = params.Rscript.value,
		args   = Box(
			ngenes  = 10,
			isTCGA  = False,
			genome  = params.genome.value,
			devpars = Box(res = 300, height = 2000, width = 2000),
			nthread = 1,
			# exclude mutation types, only consider nonsynonymous mutations.
			extypes = ["Intron", "5'UTR", "3'UTR", "IGR", "5'Flank", "3'Flank", "Silent"],
			plot    = Box(
				summary        = True,
				oncoplot       = True,
				oncostrip      = True,
				titv           = True,
				lollipop       = True,
				cbsseg         = True,
				rainfall       = True,
				tcgacomp       = True,
				vaf            = True,
				genecloud      = True,
				gisticGenome   = True,
				gisticBubble   = True,
				gisticOncoplot = True,
				somInteraction = True,
				oncodrive      = True,
				pfam           = True,
				pancan         = True,
				survival       = True,
				heterogeneity  = True,
				signature      = True,
			),
			params = Box(
				summary        = Box(rmOutlier = True, addStat = 'median', dashboard = True),
				oncoplot       = Box(),
				oncostrip      = Box(),
				titv           = Box(),
				lollipop       = Box(),
				cbsseg         = Box(labelAll = True),
				rainfall       = Box(detectChangePoints = True),
				tcgacomp       = Box(),
				vaf            = Box(flip = True),
				genecloud      = Box(minMut = 3),
				gisticGenome   = Box(markBands = 'all'),
				gisticBubble   = Box(),
				gisticOncoplot = Box(),
				somInteraction = Box(),
				oncodrive      = Box(minMut = 5, pvalMethod = 'zscore', fdrCutOff = 0.1, useFraction = True),
				pfam           = Box(),
				pancan         = Box(qval = 0.1, label = 1),
				survival       = Box(),
				heterogeneity  = Box(),
				signature      = Box(nTry = 6, plotBestFitRes = False),
			)
		)
	)

@procfactory
def _pMutationSigs():
	"""
	@name:
		pMutationSigs
	@description:
		Find similar COSMIC mutation signatures for MAF file
		using https://github.com/pwwang/deconstruct_sigs_py
	@input:
		infile: The input maf file.
	@output:
		`outdir:dir`: The output directory
	@args:
		`font_family`: Font family for plotting.
		`font_weight`: Font weight for plotting.
		`sig_cutoff` : Significance cutoff for signatures.
		`err_thres`  : The threshold to top the iteration.
		`ref`        : The reference genome.
	"""
	pMutationSigs                  = Proc(desc = 'Find similar COSMIC mutation signatures for MAF file.')
	pMutationSigs.input            = 'infile:file'
	pMutationSigs.output           = 'outdir:dir:{{i.infile | fn2}}.signature'
	pMutationSigs.args.font_family = 'Arial'
	pMutationSigs.args.font_weight = 'bold'
	pMutationSigs.args.sig_cutoff  = 0.05
	pMutationSigs.args.err_thres   = 1e-3
	pMutationSigs.args.ref         = params.ref.value
	pMutationSigs.lang             = params.python.value
	pMutationSigs.script           = "file:scripts/tcgamaf/pMutationSigs.py"
	return pMutationSigs

@procfactory
def _pSnpEff():
	"""
	@name:
		pSnpEff
	@description:
		This is the default command. It is used for annotating variant filed (e.g. VCF files).
	@input:
		infile:  The input file
	@output:
		`outdir:file`: The directory containing output anntated file, snpEff_genes.txt and snpEff_summary.html
	@args:
		`snpEff`:       The snpEff executable, default: "snpEff"
		`params`:    Other parameters for `snpEff`, default: "-Xms1g -Xmx4g -v"
		`genome`:    The genome used for annotation, default: "hg19"
		`informat`:  The format of input file [vcf or bed], default: "vcf"
		`outformat`: The format of output file [vcf, gatk, bed, bedAnn], default: "vcf"
		`csvStats`:  Whether to generate csv stats file, default: True.
		`htmlStats`: Whether to generate the html summary file, default: False.
		`javamem`:   The memory to use. Default: '-Xms1g -Xmx8g'
	@requires:
		[snpEff](http://snpeff.sourceforge.net/SnpEff_manual.html)
	"""
	pSnpEff = Proc()
	pSnpEff.input  = "infile:file"
	pSnpEff.output = "outdir:dir:{{infile | fn}}.snpeff"
	pSnpEff.args   = { "snpEff": "snpEff", "javamem": "-Xms1g -Xmx8g", "genome": "hg19", "informat": "vcf", "outformat": "vcf", "csvStats": True, "htmlStats": False, "params": "" }
	pSnpEff.script = """
	csvfile="{{outdir}}/{{infile | fn}}.csvstat"
	sumfile="{{outdir}}/{{infile | fn}}.html"
	outfile="{{outdir}}/{{infile | fn}}.snpEff.vcf"
	csvStats=""
	if [[ "{{args.csvStats}}" == "True" ]]; then
		csvStats="-csvStats \\"$csvfile\\""
	fi
	stats=""
	if [[ "{{args.htmlStats}}" == "True" ]]; then
		stats="-stats \\"$sumfile\\""
	fi
	echo {{args.snpEff}} {{args.javamem}} -i {{args.informat}} -o {{args.outformat}} $csvStats $stats {{args.params}} {{args.genome}} "{{infile}}"
	{{args.snpEff}} {{args.javamem}} -i {{args.informat}} -o {{args.outformat}} $csvStats $stats {{args.params}} {{args.genome}} "{{infile}}" > "$outfile"
	"""
	return pSnpEff

@procfactory
def _pDToxoG():
	"""
	@description:
		Run D-ToxoG on MAF files with columns listed at:
		https://software.broadinstitute.org/cancer/cga/dtoxog
		However,`i_picard_oxoQ` is required but not documented. This process
		will try to put 0's for the field to let the program run.
		To be stricter, use picard's CollectOxoGMetrics to get this column.
	@input:
		infile: The input maf file
	@output:
		outfile: The output maf file.
			- with extra columns added:
			- `pox` --  p-value that the call is actually an artifact
			- `qox` --  false detection rate score.
			- `pox_cutoff` --  minimum pox score for artifact.
			- `isArtifactMode`:
				- Variant is C>A, G>T: 1
				- Variant is not C>A or G>T: 0
			- `oxoGCut`:
				- Variant is marked as artifact: 1
				- Variant is not an artifact: 0
	@args:
		dtoxog : D-ToxoG executable.
		nthread: Maximum threads used by matlab.
		keep   : Whether keep those artifact mode mutations in output MAF file or not.
		params : Other parameters for `startFilterMAFFile`
			- See more in `startFilterMAFFile.m` or run `dtoxog` directly
	"""
	return Box(
		desc   = 'Discovery and characterization of artifactual mutations',
		input  = 'infile:file',
		output = 'outfile:file:{{i.infile | stem}}.dtoxog.maf',
		lang   = params.python.value,
		args   = Box(
			nthread = 1,
			keep    = True,
			dtoxog  = params.dtoxog.value,
			params  = Box(isGeneratingPlots = True, globalPoxoG = .96, artifactThresholdRate = .01)
		))

@procfactory
def _pMaf2Vcf():
	"""
	@input:
		infile: The input MAF file
	@output:
		outfile: Output multi-sample VCF containing all TN-pairs
		outdir: A directory with output VCF files
	@args:
		maf2vcf (str): Path to `maf2vcf.pl` from `vcf2maf` tool.
		merge: Whether merge the VCF for individual samples or not.
		params: Other parameters for `maf2vcf.pl`
	"""
	return Box(
		desc = 'Convert MAF file back to VCF files',
		lang = params.python.value,
		input = 'infile:file',
		output = [
			'outfile:file:{{i.infile | stem}}.vcfs/{{i.infile | stem}}.vcf',
			'outdir:dir:{{i.infile | stem}}.vcfs'],
		args = Box(
			maf2vcf = params.maf2vcf.value,
			ref     = params.ref.value,
			params  = Box({'per-tn-vcfs': True})),
	)


@procfactory
def _pMafAddChr():
	"""
	@input:
		infile: The input MAF file
	@output:
		outfile: The output MAF file
	"""
	return Box(
		desc = 'Add chr to chromosome if not present',
		lang = params.python.value,
		input = 'infile:file',
		output = 'outfile:file:{{i.infile | stem}}.maf'
	)

@procfactory
def _pMafExtractSample(alias = 'pMafSampleFilter'):
	"""
	@input:
		infile: The input MAF file
		samfile: The tumor sample file
			- Could also be a list of samples, separated by comma
			- Overwrite `args.samples`
	@output:
		outfile: The output file contain only the given samples
	@args:
		samples (str|list): The samples or a list of samples
	"""
	return Box(
		desc   = 'Filter MAF file with given samples',
		input  = 'infile:file, samfile:var',
		output = 'outfile:file:{{i.infile | stem}}.subset.maf',
		lang   = params.python.value,
		args   = Box(samples = [])
	)

@procfactory
def _pMafExtractClass():
	"""
	@input:
		infile: The input MAF file
		classfile: The variant class file
			- Could also be a list of classes, separated by comma
			- Overwrite `args.classes`
	@output:
		outfile: The output file contain only the given classes
	@args:
		classes (str|list): The classes or a list of classes
	"""
	return Box(
		desc   = 'Filter MAF file with given classes',
		input  = 'infile:file, classfile:var',
		output = 'outfile:file:{{i.infile | stem}}.subclass.maf',
		lang   = params.python.value,
		args   = Box(classes = [])
	)

@procfactory
def _pMafGetSamples():
	"""
	@input:
		infile: The input MAF file
	@output:
		outfile: The output sample file
	@args:
		out: Which samples to output, tumor, normal or both.
		outopts: The output options
	"""
	return Box(
		desc   = 'Get samples from MAF file',
		lang   = params.python.value,
		input  = 'infile:file',
		output = 'outfile:file:{{i.infile | stem}}.samples.txt',
		args   = Box(out = 'both', outopts = Box(cnames = True))
	)


import shlex
from os import path, rename
{{runcmd}}
{{params2CmdArgs}}

params = {{args.params}}
{% if i.infile.endswith('avinput') %}
params['annovar-file'] = {{i.infile | quote}}
{% else %}
params['vcf-file'] = {{i.infile | quote}}
{% endif %}
params['regulatory-causing-predict'] = 'all'
params['cell']                       = {{args.cell | quote}}
params['db-score']                   = 'dbncfp'
params['out']                        = {{o.outfile | prefix | prefix | quote}}

jarfile = [x for x in shlex.split({{args.cepip | quote}}) if x.endswith('.jar')][0]
jardir  = path.dirname(jarfile)

cmd = 'cd "%s"; {{args.cepip}} %s' % (jardir, params2CmdArgs(params, dash='--', equal=' '))
runcmd(cmd)

rename(params['out'] + '.flt.txt', {{o.outfile | quote}})


import re
from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell

# variables
infile1   = {{i.fq1 | quote}}
infile2   = {{i.fq2 | quote}}
outfile   = {{o.outfile | quote}}
outprefix = {{o.outfile | fn | quote}}
outfmt    = {{args.outfmt | quote}}
outdir    = {{job.outdir | quote}}
ref       = {{args.ref | quote}}
rg        = {{args.rg | repr}}
params    = {{args.params | repr}}
samtools  = {{args.samtools | quote}}
bowtie2   = {{args.bowtie2 | quote}}
bwa       = {{args.bwa | quote}}
ngm       = {{args.ngm | quote}}
star      = {{args.star | quote}}
tool      = {{args.tool | quote}}
nthread   = {{args.nthread}}
jobindex  = {{job.index}}

shell.load_config(dict(
	bowtie2  = bowtie2,
	samtools = samtools,
	bwa      = bwa,
	star     = star,
	ngm      = ngm,
))

# detemine default read group
rg = {key.upper():val for key, val in rg.items()}
if not rg['ID']:
	g = re.search (r'[^a-zA-Z0-9]+(L\\d+)[^a-zA-Z0-9]+', outprefix)
	rg['ID'] = g.group(1) if g else "{outprefix}.L{jobindex}".format(outprefix = outprefix, jobindex = jobindex)
if not rg['SM']:
	rg['SM'] = outprefix

def sam2bam(samfile, bamfile):
	shell.samtools.view(
		S = True, b = True, _out = bamfile, _ = samfile, **{'@': nthread}
	)
	shell.rm_rf(samfile)

def run_bowtie2():
	params.threads = nthread
	params.x = ref
	params.S = outfile if outfmt == 'sam' else path.splitext(outfile)[0] + '.sam'
	params['1'] = infile1
	params['2'] = infile2
	params['rg-id'] = rg['ID']
	params['rg'] = ['{}:{}'.format(k, v) for k, v in rg.items() if k != 'ID']
	shell.fg.bowtie2(**params)
	if outfmt == 'bam': sam2bam(params.S, outfile)

def run_bwa():
	params.t = nthread
	params.R = "@RG\\tID:{id}\\t{rg}".format(id = rg['ID'], rg = "\\t".join(
		'{}:{}'.format(k, v) for k, v in rg.items() if k != 'ID'
	))
	params.o = outfile if outfmt == 'sam' else path.splitext(outfile)[0] + '.sam'
	params._ = [ref, infile1, infile2]
	shell.fg.bwa.mem(**params)
	if outfmt == 'bam': sam2bam(params.o, outfile)

def run_ngm():
	params['1'] = infile1
	params['2'] = infile2
	params.r = ref
	params.o = outfile
	params.b = outfmt == 'bam'
	params.t = nthread
	for k, v in rg.items():
		params['rg-' + k.lower()] = v
	shell.fg.ngm(**params)

def run_star():
	params.genomeDir        = path.splitext(ref)[0] + '.star'
	params.readFilesIn      = [infile1, infile2]
	params.readFilesCommand = ("cat", "zcat", "bzcat")[
		1 if infile1.endswith('.gz') else 2 if infile1.endswith('.bz2') else 0
	]
	params.readNameSeparator = '.'
	params.outFileNamePrefix = outdir + '/'
	params.outSAMtype        = [outfmt.upper(), 'Unsorted']
	shell.fg.star(**params)

	starout = path.join(outdir, "Aligned.out.{}".format(outfmt))
	if path.isfile(starout):
		shell.mv(starout, outfile)

tools = dict(
	bowtie2 = run_bowtie2,
	bwa     = run_bwa,
	ngm     = run_ngm,
	star    = run_star
)

try:
	tools[tool]()
except KeyError:
	raise KeyError('Tool {!r} not supported.'.format(tool))




from pyppl import Box
from bioprocs.utils import shell
from bioprocs.utils.reference import vcfIndex

vcffile = {{ i.vcffile | quote}}
regfile = {{ i.regfile | quote}}
outfile = {{ o.outfile | quote}}
tabix   = {{ args.tabix | quote}}
params  = {{ args.params | repr}}

shell.TOOLS.tabix = tabix
vcffile = vcfIndex(vcffile, tabix)

params._ = [vcffile, regfile]
params._stdout = outfile
shell.Shell().tabix(**params).run()




from pyppl import Box
from bioprocs.utils import shell

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
tool     = {{args.tool | quote}}
gz       = {{args.gz | repr}}
bcftools = {{args.bcftools | quote}}

shell.TOOLS.bcftools = bcftools

def run_bcftools():
	bcftools        = shell.Shell(subcmd = True, equal = ' ').bcftools
	hparams         = Box()
	hparams.h       = True
	hparams._       = infile
	hparams._stdout = outfile
	bcftools.view(**hparams).run()

	qparams = Box()
	#CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT
	qparams.f        = "%CHROM\t%POS\t%ID\t%REF\t%ALT\t%QUAL\t%FILTER\t.\tGT[\t%GT]\n"
	qparams._        = infile
	qparams._stdout_ = outfile
	if gz:
		qparams.O = 'z'
	bcftools.query(**qparams).run()

tools = dict(
	bcftools = run_bcftools
)

try:
	tools[tool]()
except KeyError:
	raise ValueError('Tool {!r} not supported yet.'.format(tool))



"""Utilities of cnvkit"""
from pyppl import Proc, Box
from . import params
from .utils import fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pCNVkitPrepare():
	"""
	@name:
		pCNVkitPrepare
	@description:
		Generate target files for cnvkit, using probably cnvkit's access, target, autobin commands.
	@input:
		`infiles:files`: The bam files. Indexes are necessary.
			- Hint: if indexes are not with the input files, you probably need `pCNVkitPrepare.infile = 'origin'`
	@output:
		`target:file`:     The (autobinned) target file
		`antitarget:file`: The (autobinned) target file
	@args:
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`baits` : The bait file for the regions you captured in the experiment.
			- See https://github.com/AstraZeneca-NGS/reference_data/tree/master/hg19/bed
		`accfile`: Directly use the access file. Default: generating from the reference file.
			- See https://github.com/etal/cnvkit/tree/master/data
		`nthread`: The number of threads to use. Default: 1
		`ref`    : The reference genome.
		`params` : The extra parameters for cnvkit's `access`, `target` and `autobin` command. Default:
			```python
			Box(
				target  = Box({'short-name': True, 'split': True}),
				access  = Box(s = '5000'),
				autobin = Box()
			)
			```
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkitPrepare              = Proc(desc = 'Generate target files for cnvkit.')
	pCNVkitPrepare.input        = 'infiles:files'
	pCNVkitPrepare.output       = 'target:file:{{i.infiles | fs2name}}.target.bed, antitarget:file:{{i.infiles | fs2name}}.antitarget.bed'
	pCNVkitPrepare.args.cnvkit  = params.cnvkit.value
	pCNVkitPrepare.args.baits   = params.refexon.value
	pCNVkitPrepare.args.accfile = ''
	pCNVkitPrepare.args.nthread = 1
	pCNVkitPrepare.args.ref     = params.ref.value
	pCNVkitPrepare.args.params  = Box(
		target  = Box({'short-name': True, 'split': True}),
		access  = Box(s = '5000'),
		autobin = Box()
	)
	pCNVkitPrepare.envs.fs2name = fs2name
	pCNVkitPrepare.lang         = params.python.value
	pCNVkitPrepare.script       = 'file:scripts/cnvkit/pCNVkitPrepare.py'
	return pCNVkitPrepare

@procfactory
def _pCNVkitCov():
	"""
	@name:
		pCNVkitCov
	@description:
		Calculate coverage in the given regions from BAM read depths.
	@input:
		`infile:file`:  The bam file
		`tgfile:file`:  The target file
		`atgfile:file`: The antitarget file
	@output:
		`outfile:file`: The output cnn file
	@args:
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`nthread`: The number of threads to use. Default: 1
		`params`:  Other parameters for `cnvkit.py coverage`
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkitCov              = Proc (desc = 'Calculate coverage in the given regions from BAM read depths.')
	pCNVkitCov.input        = "infile:file, tgfile:file, atgfile:file"
	pCNVkitCov.output       = "outfile:file:{{i.infile | fn}}.target.cnn, antifile:file:{{i.infile | fn}}.antitarget.cnn"
	pCNVkitCov.args.cnvkit  = params.cnvkit.value
	pCNVkitCov.args.nthread = 1
	pCNVkitCov.args.params  = Box()
	pCNVkitCov.lang         = params.python.value
	pCNVkitCov.script       = "file:scripts/cnvkit/pCNVkitCov.py"
	return pCNVkitCov

@procfactory
def _pCNVkitRef():
	"""
	@name:
		pCNVkitRef
	@description:
		Compile a copy-number reference from the given files or directory (containing normal samples). If given a reference genome (-f option), also calculate the GC content and repeat-masked proportion of each region.
	@input:
		`infiles:files`:  The input reference coverage files
	@output:
		`outfile:file`: The output reference cnn file
	@args:
		`ref`   :  The reference file.
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`nthread`: The number of threads to use. Default: 1
		`params`:  Other parameters for `cnvkit.py reference`, default: " --no-edge "
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkitRef              = Proc (desc = 'Compile a copy-number reference from the given files or directory')
	pCNVkitRef.input        = "infiles:files"
	pCNVkitRef.output       = "outfile:file:{{i.infiles | fs2name}}.reference.cnn"
	pCNVkitRef.args.cnvkit  = params.cnvkit.value
	pCNVkitRef.args.ref     = params.ref.value
	pCNVkitRef.args.nthread = 1
	pCNVkitRef.args.params  = Box({'no-edge': True})
	pCNVkitRef.envs.fs2name = fs2name
	pCNVkitRef.lang         = params.python.value
	pCNVkitRef.script       = "file:scripts/cnvkit/pCNVkitRef.py"
	return pCNVkitRef

@procfactory
def _pCNVkitFlatRef():
	"""
	@name:
		pCNVkitFlatRef
	@description:
		Generate reference coverage if there are no normal samples.
	@input:
		`tgfile:file`:  The target file
		`atgfile:file`: The antitarget file
	@output:
		`outfile:file`: The output reference cnn file
	@args:
		`ref`   :  The reference file.
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`params`:  Other parameters for `cnvkit.py reference`, default: `{}`
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkitFlatRef              = Proc (desc = 'Compile a copy-number flat reference without normal samples')
	pCNVkitFlatRef.input        = "tgfile:file, atgfile:file"
	pCNVkitFlatRef.output       = "outfile:file:{{i.tgfile | fn}}.reference.cnn"
	pCNVkitFlatRef.args.cnvkit  = params.cnvkit.value
	pCNVkitFlatRef.args.ref     = params.ref.value
	pCNVkitFlatRef.args.params  = Box()
	pCNVkitFlatRef.envs.fs2name = fs2name
	pCNVkitFlatRef.lang         = params.python.value
	pCNVkitFlatRef.script       = "file:scripts/cnvkit/pCNVkitFlatRef.py"
	return pCNVkitFlatRef

@procfactory
def _pCNVkitFix():
	"""
	@name:
		pCNVkitFix
	@description:
		Combine the uncorrected target and antitarget coverage tables (.cnn) and correct for biases in regional coverage and GC content, according to the given reference. Output a table of copy number ratios (.cnr)
	@input:
		`tgfile:file`:  The target coverage file
		`atgfile:file`: The antitarget coverage file
		`rcfile:file`:  The reference cnn file
	@output:
		`outfile:file`: The cnr file
	@args:
		`nthread`: The number of threads to use. Default: 1
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`params`:  Other parameters for `cnvkit.py fix`, default: " --no-edge "
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkitFix              = Proc (desc = 'Combine the uncorrected target and antitarget coverage tables and correct them.')
	pCNVkitFix.input        = "tgfile:file, atgfile:file, rcfile:file"
	pCNVkitFix.output       = "outfile:file:{{i.tgfile | fn}}.cnr"
	pCNVkitFix.args.cnvkit  = params.cnvkit.value
	pCNVkitFix.args.params  = Box({'no-edge': True})
	pCNVkitFix.args.nthread = 1
	pCNVkitFix.lang         = params.python.value
	pCNVkitFix.script       = "file:scripts/cnvkit/pCNVkitFix.py"
	return pCNVkitFix

@procfactory
def _pCNVkitSeg():
	"""
	@name:
		pCNVkitSeg
	@description:
		Infer discrete copy number segments from the given coverage table
	@input:
		`infile:file`:  The cnr file
	@output:
		`outfile:file`: The cns file
	@args:
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`nthread`: The number of threads to use. Default: 1
		`params`:  Other parameters for `cnvkit.py segment`, default: ""
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkitSeg              = Proc (desc = 'Infer discrete copy number segments from the given coverage table.')
	pCNVkitSeg.input        = "infile:file"
	pCNVkitSeg.output       = "outfile:file:{{i.infile | fn}}.cns"
	pCNVkitSeg.args.cnvkit  = params.cnvkit.value
	pCNVkitSeg.args.nthread = 1
	pCNVkitSeg.args.params  = Box()
	pCNVkitSeg.lang         = params.python.value
	pCNVkitSeg.script       = "file:scripts/cnvkit/pCNVkitSeg.py"
	return pCNVkitSeg

@procfactory
def _pCNVkitCall():
	"""
	@name:
		pCNVkitCall
	@description:
		Given segmented log2 ratio estimates (.cns), derive each segment's absolute integer copy number
	@input:
		`infile:file`:  The cns file
	@output:
		`outfile:file`: The callcns file
	@args:
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`params`:  Other parameters for `cnvkit.py segment`, default: ""
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkitCall             = Proc (desc="Given segmented log2 ratio estimates (.cns), derive each segment's absolute integer copy number")
	pCNVkitCall.input       = "infile:file"
	pCNVkitCall.output      = "outfile:file:{{i.infile | fn}}.callcns"
	pCNVkitCall.args.cnvkit = params.cnvkit.value
	pCNVkitCall.args.params = Box()
	pCNVkitCall.lang        = params.python.value
	pCNVkitCall.script      = "file:scripts/cnvkit/pCNVkitCall.py"
	return pCNVkitCall

@procfactory
def _pCNVkitScatter():
	"""
	@name:
		pCNVkitScatter
	@description:
		Generate scatter plot for CNVkit results.
	@input:
		`cnrfile:file`: The cnr file
		`cnsfile:file`: The cns file from call
	@output:
		`outdir:dir`: The output directory
	@args:
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`nthread`: The number of threads to use. Default: 1
		`params`:  Other parameters for `cnvkit.py scatter`
		`regions`: The regoins to plot. Default: `['']`
			- You can have extra specific regions, format:
			- `chr5:100-50000000:TERT` or `chr7:BRAF,MET` (genes are used to highlight)
	"""
	pCNVkitScatter              = Proc(desc = 'Generate scatter plot for CNVkit results.')
	pCNVkitScatter.input        = 'cnrfile:file, cnsfile:file'
	pCNVkitScatter.output       = 'outdir:dir:{{i.cnrfile | fn}}.scatters'
	pCNVkitScatter.args.cnvkit  = params.cnvkit.value
	pCNVkitScatter.args.nthread = 1
	pCNVkitScatter.args.params  = Box()
	pCNVkitScatter.args.regions = [
		'', # plot whole genome
		# extra regions, format: chr5:100-50000000:TERT
		# or chr7:BRAF,MET
	]
	pCNVkitScatter.lang   = params.python.value
	pCNVkitScatter.script = "file:scripts/cnvkit/pCNVkitScatter.py"
	return pCNVkitScatter

@procfactory
def _pCNVkitDiagram():
	"""
	@name:
		pCNVkitDiagram
	@description:
		Generate diagram plot for CNVkit results.
	@input:
		`cnrfile:file`: The cnr file
		`cnsfile:file`: The cns file from call
	@output:
		`outfile:file`: The output file
	@args:
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`nthread`: The number of threads to use. Default: 1
		`params`:  Other parameters for `cnvkit.py scatter`
	"""
	pCNVkitDiagram             = Proc(desc = 'Generate diagram plot for CNVkit results.')
	pCNVkitDiagram.input       = 'cnrfile:file, cnsfile:file'
	pCNVkitDiagram.output      = 'outfile:file:{{i.cnrfile | fn}}.diagram.pdf'
	pCNVkitDiagram.args.cnvkit = params.cnvkit.value
	pCNVkitDiagram.args.nthread = 1
	pCNVkitDiagram.args.params = Box()
	pCNVkitDiagram.lang        = params.python.value
	pCNVkitDiagram.script      = "file:scripts/cnvkit/pCNVkitDiagram.py"
	return pCNVkitDiagram

@procfactory
def _pCNVkitHeatmap():
	"""
	@name:
		pCNVkitHeatmap
	@description:
		Generate heatmap plot for CNVkit results.
	@input:
		`cnfiles:files`: The cnr or cns files.
	@output:
		`outdir:dir`: The output directory
	@args:
		`cnvkit`:  The executable of cnvkit. Default: 'cnvkit.py'
		`params`:  Other parameters for `cnvkit.py scatter`
		`regions`: The regoins to plot. Default: `['']`
			- You can have extra specific regions, format:
			- `chr5:100-50000000` or `chr7` (genes are used to highlight)
	"""
	pCNVkitHeatmap              = Proc(desc = 'Generate heatmap plot for CNVkit results.')
	pCNVkitHeatmap.input        = 'cnfiles:files'
	pCNVkitHeatmap.output       = 'outdir:dir:{{i.cnfiles | fs2name}}.heatmaps'
	pCNVkitHeatmap.args.cnvkit  = params.cnvkit.value
	pCNVkitHeatmap.args.params  = Box()
	pCNVkitHeatmap.args.nthread = 1
	pCNVkitHeatmap.args.regions = [
		'', # plot whole genome
		# extra regions, format: chr5:100-50000000
		# or chr7
	]
	pCNVkitHeatmap.envs.fs2name = fs2name
	pCNVkitHeatmap.lang         = params.python.value
	pCNVkitHeatmap.script       = "file:scripts/cnvkit/pCNVkitHeatmap.py"
	return pCNVkitHeatmap

@procfactory
def _pCNVkitReport():
	"""
	@name:
		pCNVkitReport
	@description:
		Report CNVkit results
	@input:
		`cnrfile:file`:  The file containing copy number ratio
		`cnsfile:file`:  The file containing copy number segment
	@output:
		`outdir:dir`:   The output directory
	@args:
		`cnvkit` : The executable of cnvkit. Default: 'cnvkit.py'
		`nthread`: The number of threads to use. Default: 1
		`params` : Extra parameters to the commands.
			- `breaks`:       Whether to report breakpoints. Default: True
			- `gainloss`:     Whether to report gainloss. Default: True
			- `metrics`:      Whether to report metrics. Default: True
			- `segmetrics`:   Whether to report segmetrics. Default: True
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkitReport              = Proc (desc = 'Report CNVkit results')
	pCNVkitReport.input        = "cnrfile:file, cnsfile:file"
	pCNVkitReport.output       = "outdir:dir:{{i.cnrfile | fn}}.cnvkit.reports"
	pCNVkitReport.args.cnvkit  = params.cnvkit.value
	pCNVkitReport.args.nthread = 1
	pCNVkitReport.args.params  = Box(
		breaks     = True, # None to disable
		gainloss   = True,
		metrics    = True,
		segmetrics = Box(iqr = True)
	)
	pCNVkitReport.lang   = params.python.value
	pCNVkitReport.script = 'file:scripts/cnvkit/pCNVkitReport.py'
	return pCNVkitReport

@procfactory
def _pCNVkit2Vcf():
	"""
	@name:
		pCNVkit2Vcf
	@description:
		Output vcf file for cnvkit results
	@input:
		`cnsfile:file`: The cns file
	@output:
		`outfile:file`: The vcf file
	@args:
		`cnvkit`:   The executable of cnvkit. Default: 'cnvkit.py'
		`nthread`: The number of threads to use. Default: 1
		`params`:   Other params for `cnvkit.py export`
	@requires:
		[CNVkit](http://cnvkit.readthedocs.io/)
	"""
	pCNVkit2Vcf              = Proc (desc = 'Output vcf file for cnvkit results')
	pCNVkit2Vcf.input        = "cnsfile:file"
	pCNVkit2Vcf.output       = "outfile:file:{{i.cnsfile | fn}}.cnvkit.vcf"
	pCNVkit2Vcf.args.cnvkit  = params.cnvkit.value
	pCNVkit2Vcf.args.nthread = 1
	pCNVkit2Vcf.args.params  = Box()
	pCNVkit2Vcf.lang         = params.python.value
	pCNVkit2Vcf.script       = 'file:scripts/cnvkit/pCNVkit2Vcf.py'
	return pCNVkit2Vcf

@procfactory
def _pCNVkit2Theta():
	"""
	@name:
		pCNVkit2Theta
	@description:
		Convert the results to THetA2 interval input.
	@input:
		`cnsfile:file`: The cns file
		`cnnfile:file`: The reference cnn file or the cnr file for paired Normal sample. Could be empty.
	@output:
		`outfile:file`: The interval file for THetA2
	@args:
		`nthread` : Number threads to use. Default: `1`
		`cnvkit`  : The executable of cnvkit. Default: `cnvkit.py`
		`params`  : Other params for `cnvkit.py export theta`
	"""
	pCNVkit2Theta              = Proc(desc = 'Convert the results to THetA2 interval input.')
	pCNVkit2Theta.input        = 'cnsfile:file, cnnfile:file'
	pCNVkit2Theta.output       = 'outfile:file:{{i.cnsfile | fn2}}.interval.txt'
	pCNVkit2Theta.args.nthread = 1
	pCNVkit2Theta.args.params  = Box()
	pCNVkit2Theta.args.cnvkit  = params.cnvkit.value
	pCNVkit2Theta.lang         = params.python.value
	pCNVkit2Theta.script       = 'file:scripts/cnvkit/pCNVkit2Theta.py'
	return pCNVkit2Theta




from os import path, makedirs
from shutil import rmtree
from sys import stderr
from pyppl import Box
from bioprocs.utils import mem2, shell2 as shell
from bioprocs.utils.reference import bamIndex

{% python from os import path %}

ref       = {{args.ref | quote}}
jobindex  = {{job.index | repr}}
tmpdir    = {{args.tmpdir | quote}}
procid    = {{proc.id | quote}}
tumor     = {{i.tumor | quote}}
tprefix   = {{i.tumor | fn2 | quote}}
normal    = {{i.normal | quote}}
nprefix   = {{i.normal | fn2 | quote}}
tool      = {{args.tool | quote}}
gz        = {{args.gz | bool}}
outfile   = {{o.outfile | quote}}
params    = {{args.params | repr}}
samtools  = {{args.samtools | quote}}
mem       = {{args.mem | quote}}
nthread   = {{args.nthread | repr}}
gatk      = {{args.gatk | quote}}
ssniper   = {{args.somaticsniper | quote}}
ssniffer  = {{args.snvsniffer | quote}}
strelka   = {{args.strelka | quote}}
virmid    = {{ args.virmid | quote}}
vardict   = {{ args.vardict | quote}}
joboutdir = {{job.outdir | quote}}
bamIndex(tumor,  samtools = None, nthread = nthread)
bamIndex(normal, samtools = None, nthread = nthread)

shell.load_config(
	samtools = samtools, gatk = gatk, somaticsniper = ssniper,
	snvsniffer = ssniffer, strelka = strelka, virmid = virmid, vardict = vardict)

if tprefix == nprefix:
	tprefix = tprefix + '_TUMOR'
	nprefix = nprefix + '_NORMAL'

if gz:
	# remove ending .gz
	outfile = outfile[:-3]

if not path.exists (ref):
	stderr.write ("Reference file not specified")
	exit (1)

tmpdir = path.join (tmpdir, "{procid}.{tprefix}.{nprefix}.{jobindex}".format(
	jobindex = jobindex,
	procid   = procid,
	tprefix  = tprefix,
	nprefix  = nprefix
))
if not path.exists (tmpdir):
	makedirs (tmpdir)

def run_gatk():
	# generate interval list file
	intvfile = {{job.outdir | path.join: "interval.list" | quote}}
	shell.pipe.samtools.idxstats(tumor) | shell.pipe.head(n = 1) | shell.cut(f = 1, _out = intvfile)

	mem = mem2(mem, 'java')

	params['I:tumor']  = tumor
	params['I:normal'] = normal

	params.R   = ref
	params.o   = outfile
	params.nct = nthread
	params.L   = intvfile

	shell.fg.gatk(
		mem2(mem, 'java'),
		{'Djava.io.tmpdir': tmpdir, '_prefix': '-', '_sep': '='},
		T = 'MuTect2',
		**params)
	if gz:
		shell.gzip(outfile)

def run_somaticsniper():
	params.f   = ref
	params.F   = 'vcf'
	params[''] = [tumor, normal, outfile]
	shell.fg.somaticsniper(**params)
	if gz:
		shell.gzip(outfile)

def run_snvsniffer():
	# generate a header file
	theader = {{job.outdir | path.join: bn(i.tumor)  | @append: '.header' | quote}}
	nheader = {{job.outdir | path.join: bn(i.normal) | @append: '.header' | quote}}
	shell.samtools.view(H = True, _ = tumor, _out = theader)
	shell.samtools.view(H = True, _ = normal, _out = nheader)

	params.g = ref
	params.o = outfile

	params[''] = [theader, nheader, tumor, normal]
	shell.fg.snvsniffer.somatic(**params)
	if gz:
		shell.gzip(outfile)

def _mergeAndAddGT(snvvcf, indvcf, outfile):
	from pysam import VariantFile
	snv = VariantFile(snvvcf)
	ind = VariantFile(indvcf)

	snv.header.info.add('TYPE', 1, 'String', 'Type of somatic mutation')
	ind.header.info.add('TYPE', 1, 'String', 'Type of somatic mutation')
	snv.header.info.add('QSI', 1, 'Integer', 'Quality score for any somatic variant, ie. for the ALT haplotype to be present at a significantly different frequency in the tumor and normal')
	snv.header.info.add('TQSI', 1, 'Integer', 'Data tier used to compute QSI')
	snv.header.info.add('QSI_NT', 1, 'Integer', 'Quality score reflecting the joint probability of a somatic variant and NT')
	snv.header.info.add('TQSI_NT', 1, 'Integer', 'Data tier used to compute QSI_NT')
	snv.header.info.add('IC', 1, 'Integer', 'Number of times RU repeats in the indel allele')
	snv.header.info.add('IHP', 1, 'Integer', 'Largest reference interrupted homopolymer length intersecting with the indel')
	snv.header.info.add('OVERLAP', 0, 'Flag', 'Somatic indel possibly overlaps a second indel.')
	snv.header.info.add('RC', 1, 'Integer', 'Number of times RU repeats in the reference allele')
	snv.header.info.add('RU', 1, 'String', 'Smallest repeating sequence unit in inserted or deleted sequence')
	snv.header.formats.add('GT', 1, 'String', 'Possible genotype')
	ind.header.formats.add('GT', 1, 'String', 'Possible genotype')
	snv.header.formats.add('BCN50', 1, 'Float', 'Fraction of filtered reads within 50 bases of the indel.')
	snv.header.formats.add('DP2', 1, 'Integer', 'Read depth for tier2')
	snv.header.formats.add('DP50', 1, 'Float', 'Average tier1 read depth within 50 bases')
	snv.header.formats.add('FDP50', 1, 'Float', 'Average tier1 number of basecalls filtered from original read depth within 50 bases')
	snv.header.formats.add('SUBDP50', 1, 'Float', 'Average number of reads below tier1 mapping quality threshold aligned across sites within 50 bases')
	snv.header.formats.add('AD', 2, 'Integer', 'Allele depths for tier 1')
	snv.header.formats.add('AD2', 2, 'Integer', 'Allele depths for tier 2')
	snv.header.formats.add('TAR', 2, 'Integer', 'Reads strongly supporting alternate allele for tiers 1,2')
	snv.header.formats.add('TIR', 2, 'Integer', 'Reads strongly supporting indel allele for tiers 1,2')
	snv.header.formats.add('TOR', 2, 'Integer', 'Other reads (weak support or insufficient indel breakpoint overlap) for tiers 1,2')

	contigs = list(snv.header.contigs.keys())
	out = open(outfile, 'w')
	#Can't change sample names with VariantFile
	#out = VariantFile(outfile, 'w', header = snv.header)
	headers = str(snv.header).splitlines()
	cnames  = headers[-1].split("\t")

	cnames [-2] = nprefix
	cnames [-1] = tprefix
	headers[-1] = "\t".join(cnames)
	out.write("\n".join(headers) + "\n")

	r1 = r2 = None
	indel_gts = {
		'ref': (0, 0),
		'het': (0, 1),
		'hom': (1, 1)
	}
	while True:
		if not r1:
			try:
				r1 = next(snv)
				r1.info['TYPE'] = 'SNV'
				alleles = (r1.ref, ) + r1.alts
				gts = r1.info['SGT'].split('->')
				try:
					r1.samples['NORMAL']['GT'] = tuple(sorted(alleles.index(gt)
						for gt in list(gts[0])))
					r1.samples['TUMOR']['GT'] = tuple(sorted(alleles.index(gt)
						for gt in list(gts[1])))
					assert len(r1.samples['TUMOR']['GT']) == 2
					U1 = alleles[r1.samples['TUMOR']['GT'][0]] + 'U'
					U2 = alleles[r1.samples['TUMOR']['GT'][1]] + 'U'
					r1.samples['TUMOR']['AD'] = (
						r1.samples['TUMOR'][U1][0],
						r1.samples['TUMOR'][U2][0])
					r1.samples['TUMOR']['AD2'] = (
						r1.samples['TUMOR'][U1][1],
						r1.samples['TUMOR'][U2][1])
				except ValueError:
					r1 = None
					continue
			except StopIteration:
				r1 = None
		if not r2:
			try:
				r2 = next(ind)
				r2.info['TYPE'] = 'INDEL'
				gts = r2.info['SGT'].split('->')
				r2.samples['NORMAL']['GT'] = indel_gts[gts[0]]
				r2.samples['TUMOR']['GT']  = indel_gts[gts[1]]
			except StopIteration:
				r2 = None

		if r1 and r2:
			if (contigs.index(r1.chrom), r1.pos) < (contigs.index(r2.chrom), r2.pos):
				out.write(str(r1))
				r1 = None
			else:
				out.write(str(r2))
				r2 = None
		elif r1:
			out.write(str(r1))
			r1 = None
		elif r2:
			out.write(str(r2))
			r2 = None
		else:
			break
	out.close()

def run_strelka():
	cparams                = {{args.configParams | repr}}
	cparams.normalBam      = normal
	cparams.tumorBam       = tumor
	cparams.referenceFasta = ref
	cparams.runDir         = joboutdir
	shell.fg.strelka(**cparams)

	params.m    = 'local'
	params.g    = int(float(mem2(mem, 'G')[:-1]))
	params.j    = nthread
	params._exe = path.join(joboutdir, 'runWorkflow.py')
	shell.fg.runWorkflow(**params)

	snvvcf = path.join(joboutdir, 'results', 'variants', 'somatic.snvs.vcf.gz')
	indvcf = path.join(joboutdir, 'results', 'variants', 'somatic.indels.vcf.gz')
	_mergeAndAddGT(snvvcf, indvcf, outfile)
	if gz:
		shell.gzip(outfile)

def run_virmid():
	params.R = ref
	params.D = tumor
	params.N = normal
	params.w = joboutdir
	shell.fg.virmid(
		mem2(mem, 'java'),
		{'Djava.io.tmpdir': tmpdir, '_prefix': '-', '_sep': '='},
		**params
	)

	shell.mv(path.join(joboutdir, '*.virmid.som.passed.vcf'), outfile)
	if gz:
		shell.gzip(outfile)

def run_vardict():
	params.v    = True
	params.G    = ref
	params.b    = '{}|{}'.format(tumor, normal)
	params._out = outfile
	shell.fg.vardict(**params)

	if gz:
		shell.gzip(outfile)

tools = {
	'gatk'         : run_gatk,
	'somaticsniper': run_somaticsniper,
	'snvsniffer'   : run_snvsniffer,
	'strelka'      : run_strelka,
	'virmid'       : run_virmid,
	'vardict'      : run_vardict
}

try:
	tools[tool]()
except Exception:
	raise
finally:
	rmtree (tmpdir)



from os import path
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

maffiles = {{i.infiles | repr}}
excols   = {{args.excols | quote}}
outfile  = {{o.outfile | quote}}

# first 34 columns are TCGA standard columns
headers = ["Hugo_Symbol", "Entrez_Gene_Id", "Center", "NCBI_Build", "Chromosome", "Start_Position", "End_Position", "Strand", "Variant_Classification", "Variant_Type", "Reference_Allele", "Tumor_Seq_Allele1", "Tumor_Seq_Allele2", "dbSNP_RS", "dbSNP_Val_Status", "Tumor_Sample_Barcode", "Matched_Norm_Sample_Barcode", "Match_Norm_Seq_Allele1", "Match_Norm_Seq_Allele2", "Tumor_Validation_Allele1", "Tumor_Validation_Allele2", "Match_Norm_Validation_Allele1", "Match_Norm_Validation_Allele2", "Verification_Status", "Validation_Status", "Mutation_Status", "Sequencing_Phase", "Sequence_Source", "Validation_Method", "Score", "BAM_File", "Sequencer", "Tumor_Sample_UUID", "Matched_Norm_Sample_UUID"]

# get all headers
cols   = []
for maffile in maffiles:
	reader = TsvReader(maffile)
	headers.extend(cname for cname in reader.cnames if cname not in headers)
	reader.close()

if excols == 'discard':
	headers = headers[:34]

writer = TsvWriter(outfile)
writer.cnames = headers
writer.writeHead()
for maffile in maffiles:
	reader = TsvReader(maffile)
	for r in reader:
		for head in headers:
			if head not in r:
				r[head] = '__UNKNOWN__'
		writer.write(r)
	reader.close()
writer.close()


import re
from os import path
from pyppl import Box
from bioprocs.utils import runcmd, cmdargs, logger
from bioprocs.utils.meme import MemeReader
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter, TsvRecord
from bioprocs.utils.parallel import Parallel

tffile   = {{i.tffile | quote}}
sfile    = {{i.sfile | quote}}
outfile  = {{o.outfile | quote}}
outdir   = {{o.outdir | quote}}
tool     = {{args.tool | quote}}
meme     = {{args.meme | quote}}
params   = {{args.params | repr}}
tfmotifs = {{args.tfmotifs | quote}}
pval     = {{args.pval | repr}}
ucsclink = {{args.ucsclink | quote}}
nthread  = {{args.nthread | repr}}

# get all motifs
logger.info('Loading motif names ...')
reader = TsvReader(tffile, cnames = False)
ncol   = len(next(reader))
reader.rewind()
if ncol == 1:
	motifns = {r[0]:r[0] for r in reader}
else:
	motifns = {r[0]:r[1] for r in reader}
logger.info('%s motif names read.', len(motifns))

# match motifs
logger.info('Matching motifs in database ...')
mnames = [m.name for m in MemeReader(tfmotifs)]
motifs = {k:v for k, v in motifns.items() if k in mnames}
logger.info('%s motifs loaded', len(motifs))

if tool == 'meme':
	cmdparams        = []
	params.thresh    = pval
	params.verbosity = 4
	for motif, name in motifs.items():
		params.oc    = path.join(outdir, name + '.' + re.sub(r'[^\w_]', '', motif))
		params.motif = motif
		params._     = [tfmotifs, sfile]
		cmdparams.append((meme, cmdargs(params, dash = '--', equal = ' ')))
	Parallel(nthread, raiseExc = True).run('{} {}', cmdparams)

	writer = TsvWriter(outfile)
	writer.cnames = [
		"CHR", "START", "END", "NAME", "SCORE", "STRAND", "MOTIF", "SEQ", "STARTONSEQ",
		"STOPONSEQ", "RAWSCORE", "PVAL", "QVAL", "MATCHEDSEQ", "UCSCLINK"
	]
	writer.writeHead(callback = lambda cnames: "#" + "\t".join(cnames))

	def rowfactory(r):
		if 'p-value' not in r:
			return None
		r.PVAL       = float(r['p-value'])
		if r.PVAL >= pval:
			return None
		r.RAWSCORE = r.score
		try:
			r.SCORE = int(float(r.score) * 10)
		except TypeError:
			r.SCORE = 0
		r.STRAND   = r.strand
		r.MOTIF    = r.motif_id
		# split motif_alt_id
		# GENE or GENE::chr1:111-222 or ::chr1:111-222 or chr1:111-222
		r.SEQ = r.sequence_name
		if '::' in r.SEQ:
			seqname, seqcoord = r.SEQ.split('::', 1)
			r.CHR, start, end = re.split(r'[:-]', seqcoord)
			start = int(start)
			end   = int(end)
		elif ':' in r.SEQ:
			seqname, seqcoord = r.SEQ, r.SEQ
			r.CHR, start, end = re.split(r'[:-]', seqcoord)
			start = int(start)
			end   = int(end)
		else:
			seqname, seqcoord = r.SEQ, ''
			r.CHR, start, end = '-', 0, 0
		r.NAME       = seqname
		r.STARTONSEQ = r.start
		r.STOPONSEQ  = r.stop
		r.START      = start + int(r.start)
		r.END        = start + int(r.stop)
		r.QVAL       = r['q-value']
		r.MATCHEDSEQ = r.matched_sequence
		r.UCSCLINK   = ucsclink.format('{chrom}:{start}:{end}'.format(
			chrom = r.CHR,
			start = start,
			end   = end
		))
		return r

	for motif, name in motifs.items():
		retfile = path.join(outdir, name + '.' + re.sub(r'[^\w_]', '', motif), 'fimo.tsv')
		# motif_id	motif_alt_id	sequence_name	start	stop	strand	score	p-value	q-value	matched_sequence
		logger.info('- Merging %s ...', retfile)
		reader = TsvReader(
			retfile, 
			#cnames  = lambda header: header[1:].strip().split('\t'),
			cnames  = True,
			row     = rowfactory,
			comment = None)
		for r in reader:
			if not r:
				continue
			r.NAME   = name + '::' + r.NAME
			writer.write(r)
	writer.close()
else:
	raise ValueError('Unsupported tool: {}'.format(tool))







{{dsnparse}}

from sys import stderr
dsn = dsnparse({{i.dsn | quote}})
if not hasattr(dsn, 'scheme'):
	raise Exception('Cannot determine the scheme from DSN string.')

if dsn.scheme in ['sqlite', 'sqlite3']:
	assert hasattr(dsn, 'file') and dsn.file != ':memory:'
	import sqlite3
	conn = sqlite3.connect(dsn.file)
	conn.text_factory = str
	for sql in {{args.sql}}:
		conn.execute(sql)
		stderr.write('Executing SQL:\n')
		stderr.write(createsql + '\n')
	conn.commit()
	conn.close()
else:
	raise NotImplementedError('Only support sqlite3 currently.')


"""
Reader and writer for tsv file.
"""
import sys, inspect
from pyppl import Box
from bioprocs.utils import alwaysList
from collections import OrderedDict

def _getargs(args, func):
	argnames = inspect.getargspec(func).args
	return {k:v for k, v in args.items() if k in argnames}

__all__ = ['TsvMeta', 'TsvRecord', 'TsvReader', 'TsvWriter', 'tsvops']

class NoSuchReader(Exception):
	pass

class NoSuchWriter(Exception):
	pass

class TsvRecord(object):
	__slots__ = ('__keys', '__vals')

	def __init__(self, *items, **kwargs):
		self.__keys = []
		self.__vals = []
		self.fromItems(*(list(items) + kwargs.items()))

	def fromKeyVals(self, keys = None, values = None):
		keys = keys or []
		vals = values or [None] * len(keys)
		if not isinstance(vals, list):
			vals = [vals] * len(keys)
		# avoid TsvRecord operation change original keys
		self.__keys = keys[:]
		self.__vals = vals

		for key in self.__keys:
			if str(key).startswith('__') or str(key).startswith('_TsvRecord'):
				raise KeyError('No keys startswith underscore(__) allowed.')
		# Ensure that lengths match properly.
		assert len(self.__keys) == len(self.__vals)
		# Ensure the keys are unique
		assert len(set(self.__keys)) == len(self.__keys)

	def fromItems(self, *items):
		for item in items:
			if isinstance(item, list):
				self.fromItems(*item)
			elif isinstance(item, (dict, TsvRecord)):
				self.fromItems(*item.items())
			else:
				assert isinstance(item, tuple)
				self.__setattr__(*item)

	def keys(self):
		"""Returns the list of column names from the query."""
		return self.__keys

	def values(self):
		"""Returns the list of values from the query."""
		return self.__vals

	def __repr__(self):
		return 'TsvRecord({})'.format(list(self.items()))

	def items(self):
		return ((key, val) for key, val in zip(self.__keys, self.__vals))

	def __getitem__(self, key):
		# Support for index-based lookup.
		if isinstance(key, int):
			return self.__vals[key]

		# Support for string-based lookup.
		if key in self.__keys:
			i = self.__keys.index(key)
			return self.__vals[i]

		raise KeyError("Record contains no '{}' field.".format(key))

	def __getattr__(self, key):
		if str(key).startswith('__') or str(key).startswith('_TsvRecord'):
			return super(TsvRecord, self).__getattr__(key)
		return self[key]

	def __setitem__(self, key, value):
		if isinstance(key, int):
			if key >= len(self):
				raise IndexError('Index out of range: {}'.format(key))
			self.__vals[key] = value
		elif key in self.__keys:
			i = self.__keys.index(key)
			self.__vals[i] = value
		else:
			self.__keys.append(key)
			self.__vals.append(value)

	def __setattr__(self, key, value):
		if str(key).startswith('__') or str(key).startswith('_TsvRecord'):
			super(TsvRecord, self).__setattr__(key, value)
		else:
		  self[key] = value

	def __len__(self):
		return len(self.__keys)

	def __eq__(self, other):
		return self.keys() == other.keys() and self.values() == other.values()

	def __ne__(self, other):
		return not self.__eq__(other)

	def get(self, key, default=None):
		"""Returns the value for a given key, or default."""
		try:
			return self[key]
		except KeyError:
			return default

	def asDict(self, ordered=False):
		"""Returns the row as a dictionary, as ordered."""
		items = zip(self.__keys, self.__vals)

		return OrderedDict(items) if ordered else dict(items)

	def update(self, *others, **kwargs):
		self.fromItems(*(list(others) + kwargs.items()))

	def __contains__(self, key):
		return key in self.__keys

	def __index__(self, key):
		return self.__keys.index(key)

	def __delitem__(self, key):
		i = self.__keys.index(key)
		del self.__keys[i]
		del self.__vals[i]

class TsvMeta(TsvRecord):
	"""
	Tsv meta data
	"""
	def __init__(self, *args, **kwargs):
		"""
		arg could be a string, a tuple or a list:
		'a', ('b': int) or ['c', 'd']
		"""
		super(TsvMeta, self).__init__()
		self.add(*args)

	def __repr__(self):
		return 'TsvMeta(%s)' % ', '.join(['%s=%s'%(k,v.__name__ if v else v) for k,v in self.items()])

	def add(self, *args):
		for arg in args:
			if isinstance(arg, tuple):
				if arg[1] and not callable(arg[1]):
					raise ValueError('Expect callable as meta value.')
				self.__setattr__(*arg)
			elif isinstance(arg, list):
				for a in arg:
					if a[1] and not callable(a[1]):
						raise ValueError('Expect callable as meta value.')
						self.__setattr__(*a)
					else:
						self[a] = None
			else:
				self[arg] = None

	def append(self, *args):
		self.add(*args)

	def prepend(self, *args):
		keys = []
		vals = []
		for arg in args:
			if isinstance(arg, tuple):
				if arg[1] and not callable(arg[1]):
					raise ValueError('Expect callable as meta value.')
				keys.append(arg[0])
				vals.append(arg[1])
			elif isinstance(arg, list):
				for a in arg:
					if a[1] and not callable(a[1]):
						raise ValueError('Expect callable as meta value.')
						keys.append(a[0])
						vals.append(a[1])
					else:
						keys.append(a)
						vals.append(None)
			else:
				keys.append(arg)
				vals.append(None)
		self.fromKeyVals(keys + self.keys(), vals + self.values())

class TsvReaderBase(object):
	def __init__(self, infile, delimit = '\t', comment = '#', skip = 0):
		openfunc = open
		if infile.endswith('.gz'):
			import gzip
			openfunc = gzip.open

		self.meta    = TsvMeta()
		self.file    = openfunc(infile)
		self.delimit = delimit
		self.comment = comment
		self.tell    = 0

		while True:
			tell = self.file.tell()
			line = self.file.readline()
			if self.comment and line.startswith(self.comment):
				continue
			self.file.seek(tell)
			break

		if skip > 0:
			for _ in range(skip):
				self.file.readline()
		self.tell = self.file.tell()

	def autoMeta(self, prefix = 'COL'):
		line = self.file.readline()
		while self.comment and line.startswith(self.comment):
			line = self.file.readline()
		self.rewind()
		line = line.rstrip('\n').split(self.delimit)
		cols = [prefix + str(i+1) for i in range(len(line))]
		self.meta.add(*cols)

	def _parse(self, line):
		record = TsvRecord()
		keys = self.meta.keys()
		record.fromKeyVals(keys, [
			'' if i >= len(line) else self.meta[key](line[i]) if self.meta[key] else line[i] for i, key in enumerate(keys)
		])
		return record

	def next(self):
		line = self.file.readline()
		while self.comment and line.startswith(self.comment):
			line = self.file.readline()
		line = line.rstrip('\n')
		# empty lines not allowed
		if not line: raise StopIteration()
		return self._parse(line.split(self.delimit))

	def dump(self):
		return [r for r in self]

	def rewind(self):
		self.file.seek(self.tell)

	def __iter__(self):
		return self

	def __del__(self):
		self.close()

	def close(self):
		if self.file:
			self.file.close()

class TsvReaderBed(TsvReaderBase):
	META = [
		('CHR'   , None),
		('START' , int),
		('END'   , int),
		('NAME'  , None),
		('SCORE' , float),
		('STRAND', None)
	]

	def __init__(self, infile, skip = 0, comment = '#', delimit = '\t'):
		super(TsvReaderBed, self).__init__(infile, skip = skip, comment = comment, delimit = delimit)
		self.meta = TsvMeta(*TsvReaderBed.META)
		self.index = 1

	def _parse(self, line):
		self.meta.SCORE = str
		r = super(TsvReaderBed, self)._parse(line)
		if not r.NAME:
			r.NAME = 'BED' + str(self.index)
			self.index += 1
		if not r.SCORE or r.SCORE == '.': r.SCORE   = 0.0
		if not r.STRAND: r.STRAND = '+'
		return r

class TsvReaderBed12(TsvReaderBase):
	META = [
		('CHR'         , None),
		('START'       , int),
		('END'         , int),
		('NAME'        , None),
		('SCORE'       , float),
		('STRAND'      , None),
		('THICKSTART'  , int),
		('THICKEND'    , int),
		('ITEMRGB'     , None),
		('BLOCKCOUNT'  , int),
		('BLOCKSIZES'  , None),
		('BLOCKSTARTS' , None)
	]

	def __init__(self, infile, skip = 0, comment = '#', delimit = '\t'):
		super(TsvReaderBed12, self).__init__(infile, skip = skip, comment = comment, delimit = delimit)
		self.meta = TsvMeta(*TsvReaderBed12.META)

class TsvReaderBedpe(TsvReaderBase):
	META = [
		('CHR1'    , None),
		('START1'  , int),
		('END1'    , int),
		('CHR2'    , None),
		('START2'  , int),
		('END2'    , int),
		('NAME'    , None),
		('SCORE'   , float),
		('STRAND1' , None),
		('STRAND2' , None)
	]

	def __init__(self, infile, skip = 0, comment = '#', delimit = '\t'):
		super(TsvReaderBedpe, self).__init__(infile, skip = skip, comment = comment, delimit = delimit)
		self.meta = TsvMeta(*TsvReaderBedpe.META)

class TsvReaderBedx(TsvReaderBase):
	META = [
		('CHR'   , None),
		('START' , int),
		('END'   , int),
		('NAME'  , None),
		('SCORE' , float),
		('STRAND', None)
	]

	def __init__(self, infile, skip = 0, comment = '#', delimit = '\t', xcols = None, headprefix = ''):
		super(TsvReaderBedx, self).__init__(infile, skip = skip, comment = comment, delimit = delimit)
		self.meta = TsvMeta(*TsvReaderBedx.META)

		xmeta = OrderedDict()
		if not xcols:
			pass
		elif isinstance(xcols, list):
			for xcol in xcols:
				xmeta[xcol] = None
		elif isinstance(xcols, dict):
			for xcol, callback in xcols.items():
				if not callable(callback):
					raise TypeError('Expect callable for xcols values.')
				xmeta[xcol] = callback
		else:
			xmeta[xcols] = None
		self.meta.add(*xmeta.items())

class TsvReaderHead(TsvReaderBase):

	def __init__(self, infile, comment = '#', delimit = '\t', skip = 0, tmeta = None):
		super(TsvReaderHead, self).__init__(infile, skip = skip, comment = comment, delimit = delimit)
		self.meta = TsvMeta()

		header = self.file.readline().strip('#\t\n ').split(delimit)
		self.tell = self.file.tell()
		row1   = self.file.readline().strip().split(delimit)
		if len(row1) == len(header) + 1:
			header.insert(0, 'ROWNAMES')
		metatype = OrderedDict()
		for head in header:
			metatype[head] = None if not tmeta or not isinstance(tmeta, dict) or not head in tmeta or not callable(tmeta[head]) else tmeta[head]
		self.meta.add(*metatype.items())

		self.rewind()

class TsvReaderNometa(TsvReaderHead):

	def __init__(self, infile, comment = '#', delimit = '\t', skip = 0, head = True, tmeta = None):
		if head:
			super(TsvReaderNometa, self).__init__(infile, skip = skip, comment = comment, delimit = delimit, tmeta = tmeta)
		else:
			super(TsvReaderHead, self).__init__(infile, skip = skip, comment = comment, delimit = delimit)

	def _parse(self, line):
		return line

class TsvWriterBase(object):
	def __init__(self, outfile, delimit = '\t'):
		openfunc = open
		if outfile.endswith('.gz'):
			import gzip
			openfunc = gzip.open

		self.delimit = delimit
		self.meta    = TsvMeta()
		self.file    = open(outfile, 'w')

	def writeHead(self, prefix = '', delimit = None, transform = None):
		delimit = delimit or self.delimit
		keys = self.meta.keys()
		if callable(transform):
			keys = transform(keys)
		elif isinstance(transform, dict):
			keys = [key if not key in transform or not callable(transform[key]) else transform[key](key) for key in keys]
		self.file.write(prefix + delimit.join(keys) + '\n')

	def write(self, record, delimit = None):
		delimit = delimit or self.delimit
		outs = []
		for i, key in enumerate(self.meta.keys()):
			try:
				outs.append(str(record[key]))
			except TypeError:
				outs.append(str(record[i]))

		self.file.write(delimit.join(outs) + '\n')

	def __del__(self):
		self.close()

	def close(self):
		if self.file:
			self.file.close()

class TsvWriterNometa(TsvWriterBase):

	def writeHead(self, prefix = '', delimit = None, transform = None):
		transform = transform or (lambda keys: [str(key) for key in keys])
		super(TsvWriterNometa, self).writeHead(prefix, delimit, transform)

	def write(self, record, delimit = None):
		delimit = delimit or self.delimit
		self.file.write(delimit.join([str(r) for r in record]) + '\n')

class TsvWriterBed(TsvWriterBase):

	def __init__(self, outfile, delimit = '\t'):
		super(TsvWriterBed, self).__init__(outfile)
		self.meta = TsvMeta(*TsvReaderBed.META)

class TsvReader(object):
	# inopts:
	# - delimit, comment, skip
	# - ftype
	# - cnames
	def __new__(cls, infile, **inopts):
		inopts2 = {'delimit': '\t', 'comment': '#', 'skip': 0, 'ftype': '', 'cnames': ''}
		inopts2.update(inopts)
		inopts = inopts2
		ftype  = inopts['ftype']
		cnames = inopts['cnames']
		if not isinstance(cnames, dict):
			cnames = alwaysList(cnames)
		del inopts['ftype']
		del inopts['cnames']

		if not ftype:
			inopts = _getargs(inopts, TsvReaderBase.__init__)
			reader = TsvReaderBase(infile, **inopts)
		else:
			klass = 'TsvReader' + ftype[0].upper() + ftype[1:].lower()
			if not klass in globals():
				raise NoSuchReader(klass)
			klass  = globals()[klass]
			inopts = _getargs(inopts, klass.__init__)
			reader = klass(infile, **inopts)
		if cnames:
			metas = cnames if isinstance(cnames, list) else cnames.items()
			reader.meta.add(*metas)
		return reader


class TsvWriter(object):
	def __new__(cls, outfile, **outopts):
		outopts2 = {'delimit': '\t', 'ftype': '', 'cnames': ''}
		outopts2.update(outopts)
		outopts  = outopts2
		ftype    = outopts['ftype']
		cnames   = outopts['cnames']
		#print cnames, '3242342'
		if not isinstance(cnames, dict):
			cnames = alwaysList(cnames)
		del outopts['ftype']
		del outopts['cnames']

		if not ftype:
			outopts = _getargs(outopts, TsvWriterBase.__init__)
			writer = TsvWriterBase(outfile, **outopts)
		else:
			klass = 'TsvWriter' + ftype[0].upper() + ftype[1:].lower()
			if not klass in globals():
				raise NoSuchWriter(klass)
			klass = globals()[klass]
			outopts = _getargs(outopts, klass.__init__)
			writer = klass(outfile, **outopts)
		if cnames:
			metas = cnames if isinstance(cnames, list) else cnames.items()
			writer.meta.add(*metas)
		return writer

class SimRead (object):

	def __init__(self, *files, **kwargs):

		length = len(files)

		self.do      = None
		self.comment = ["#"] * length
		self.delimit = ["\t"] * length
		self.skip    = [0] * length
		self.debug   = kwargs['debug'] if 'debug' in kwargs else False
		self.ftype   = [''] * length
		self.cnames  = [''] * length
		self.head    = [None] * length # for nometa

		if 'match' in kwargs:
			self.match = kwargs['match']
		if 'do' in kwargs:
			self.do    = kwargs['do']
		if 'delimit' in kwargs and kwargs['delimit']:
			if not isinstance(kwargs['delimit'], (tuple, list)):
				self.delimit = [kwargs['delimit']] * length
			else:
				self.delimit[:len(kwargs['delimit'])] = list(kwargs['delimit'])
		if 'skip' in kwargs and kwargs['skip']:
			if not isinstance(kwargs['skip'], (tuple, list)):
				self.skip = [kwargs['skip']] * length
			else:
				self.skip[:len(kwargs['skip'])] = list(kwargs['skip'])
		if 'comment' in kwargs and kwargs['comment']:
			if not isinstance(kwargs['comment'], (tuple, list)):
				self.comment = [kwargs['comment']] * length
			else:
				self.comment[:len(kwargs['comment'])] = list(kwargs['comment'])
		if 'ftype' in kwargs and kwargs['ftype']:
			if not isinstance(kwargs['ftype'], (tuple, list)):
				self.ftype = [kwargs['ftype']] * length
			else:
				self.ftype[:len(kwargs['ftype'])] = list(kwargs['ftype'])
		if 'cnames' in kwargs and kwargs['cnames']:
			if not isinstance(kwargs['cnames'], (tuple, list)):
				self.cnames = [kwargs['cnames']] * length
			else:
				self.cnames[:len(kwargs['cnames'])] = list(kwargs['cnames'])
		if 'head' in kwargs and kwargs['head']:
			if not isinstance(kwargs['head'], (tuple, list)):
				self.head = [kwargs['head']] * length
			else:
				self.head[:len(kwargs['head'])] = list(kwargs['head'])

		self.match = SimRead._defaultMatch

		self.readers = []
		for i, fn in enumerate(files):
			if self.ftype[i] == 'nometa':
				inopts = {
					'ftype'  : self.ftype[i],
					'cnames' : self.cnames[i],
					'delimit': self.delimit[i],
					'skip'   : self.skip[i],
					'comment': self.comment[i],
					'head'   : self.head[i]
				}
			else:
				inopts = {
					'ftype'  : self.ftype[i],
					'cnames' : self.cnames[i],
					'delimit': self.delimit[i],
					'skip'   : self.skip[i],
					'comment': self.comment[i]
				}
			reader = TsvReader(fn, **inopts)
			if not reader.meta: reader.autoMeta()
			self.readers.append(reader)

	@staticmethod
	def compare(a, b, reverse = False):
		if not reverse:
			return 0 if a < b else 1 if a > b else -1
		else:
			return 0 if a > b else 1 if a < b else -1

	@staticmethod
	def _defaultMatch(*lines):
		data = [line[0] for line in lines]
		mind = min(data)
		if data.count(mind) == len(lines):
			return -1
		else:
			return data.index(mind)

	def run (self):
		if not self.do:
			raise AttributeError('You would like to do something when lines are matched.')
		try:
			lines = [next(reader) for reader in self.readers]
		except StopIteration:
			return
		if self.debug:
			sys.stderr.write('- Lines initiated ...\n')

		while True:
			if self.debug:
				sys.stderr.write('\n'.join([('  > FILE %s: [' % (i+1)) + str(line) + ']' for i, line in enumerate(lines)]) + '\n')
			if not all(lines): break

			try:
				m = self.match(*lines)
			except Exception as ex:
				msgs = [str(ex) + ' in MATCH function:']
				for k, line in enumerate(lines):
					msgs.append('File %s: %s' % (k+1, line))
				raise type(ex)('\n'.join(msgs) + '\n')
			if self.debug:
				sys.stderr.write('  Match returns: %s\n' % m)
			if m < 0:
				if self.debug:
					sys.stderr.write('  All lines matched, do stuff.\n')
				try:
					self.do(*lines)
				except Exception as ex:
					msgs = [str(ex) + ' in DO function:']
					for k, line in enumerate(lines):
						msgs.append('File %s: %s' % (k+1, line))
					raise type(ex)('\n'.join(msgs) + '\n')
				m = 0
			if self.debug:
				sys.stderr.write('- File %s is behind, read it ...\n' % (m+1))
			try:
				lines[m] = next(self.readers[m])
			except StopIteration:
				break

def tsvops(infile, outfile, inopts, outopts, ops = None):

	inopts2 = Box(delimit = '\t', comment = '#', skip = 0, ftype = '', cnames = '')
	inopts2.update(inopts)
	inopts = inopts2

	outopts2 = Box(delimit = '\t', headPrefix = '', headDelimit = '\t', headTransform = None, head = True, ftype = '', cnames = '')
	outopts2.update(outopts)
	outopts = outopts2

	inftype       = inopts['ftype']
	outftype      = outopts['ftype']
	head          = outopts['head']
	headPrefix    = outopts['headPrefix']
	headDelimit   = outopts['headDelimit']
	headTransform = outopts['headTransform']
	del outopts['head']
	del outopts['headPrefix']
	del outopts['headDelimit']
	del outopts['headTransform']

	reader = TsvReader(infile, **inopts)
	if not reader.meta: reader.autoMeta()

	if not outftype and not outopts['cnames']:
		outftype = 'reader'

	if outftype == 'reader':
		del outopts['ftype']
		writer = TsvWriter(outfile, **outopts)
		writer.meta.prepend(*reader.meta.items())
	else:
		writer = TsvWriter(outfile, **outopts)

	if head:
		writer.writeHead(prefix = headPrefix, delimit = headDelimit, transform = headTransform)

	for r in reader:
		if callable(ops):
			r = ops(r)
			if not r: continue
		writer.write(r)




from os import path
from pyppl import ProcSet, Channel
from pyppl.procset import Values
from bioprocs.fastx import pFastqTrim, pFastq2Sam, pFastQC, pFastMC, pFastqSETrim, pFastqSE2Sam
from bioprocs.sambam import pBam2Fastq, pSam2Bam, pBam2Counts, pBamRecal, pBam2Gmut, pBamPair2Smut
from bioprocs.common import pFile2Proc, pFiles2Dir
from bioprocs.vcf import pVcf2Maf
from bioprocs.tcgamaf import pMafMerge
from bioprocs.cnvkit import pCNVkit2Vcf, pCNVkitCall, pCNVkitRef, pCNVkitFlatRef, pCNVkitFix, pCNVkitCov, pCNVkitHeatmap, pCNVkitDiagram, pCNVkitReport, pCNVkitScatter, pCNVkitSeg, pCNVkitPrepare
from bioprocs.utils.sampleinfo import SampleInfo

"""
@name:
	aPrepareBam
@description:
	Prepare call-ready bam files from external bams or fastq files
@procs:
	`pBam2Fastq`: Convert external bam files into fastq files
	`pFastqTrim`: Trim fastq reads.
	`pFastQC`   : Run fastqc
	`pFastMC`   : Run multiqc
	`pFastq2Sam`: Convert fastq files into sam/bam files.
	`pSam2Bam`  : Convert sam files into bam files and do some cleaning (sort, markdup)
	`pBamRecal` : Recalibrate bam file.
@modules
	`ebam` : The aggregation starts with external bam files (off by default)
		- `pBam2Fastq[*]` -> `pFastqTrim` -> ...
	`fastq`: The aggregation starts with external fastq files (on by default, mutually exclusive with `ebam`)
		- `pFastqTrim[*]` -> ...
	`qc`   : Do qc for the fastq files (off by default)
		- `pFastqTrim` -> `pFastQC, pFastMC`
"""
aPrepareBam = ProcSet(
	pBam2Fastq,
	pFastqTrim,
	pFastQC,
	pFastMC,
	pFastq2Sam,
	pSam2Bam,
	pBamRecal,
	depends = False
)
# starts and ends
aPrepareBam.ends = 'pBamRecal'
# depends
aPrepareBam.pFastq2Sam.depends = aPrepareBam.pFastqTrim
aPrepareBam.pSam2Bam.depends   = aPrepareBam.pFastq2Sam
aPrepareBam.pBamRecal.depends  = aPrepareBam.pSam2Bam
# delegates
aPrepareBam.delegate('args.ref', 'pFastq2Sam', 'pBamRecal')
aPrepareBam.delegate('args.gz', 'pBam2Fastq', 'pFastqTrim')
# args
aPrepareBam.pSam2Bam.args.markdup = True
# modules

@aPrepareBam.module
def aPrepareBam_ebam(ps, restore = True):
	ps.starts = 'pBam2Fastq'
	ps.pFastqTrim.depends = ps.pBam2Fastq

@aPrepareBam.module
def aPrepareBam_fastq(ps, restore = True):
	ps.starts = 'pFastqTrim'

@aPrepareBam.module
def aPrepareBam_qc(ps, restore = True):
	ps.ends = 'pFast?C'
	ps['pFast?C'].depends = ps.pFastqTrim

@aPrepareBam.module
def aPrepareBam_norecal(ps, restore = True):
	ps.ends = 'pSam2Bam'
	ps.pBamRecal.depends = []

# inital modules
aPrepareBam.modules.ebam()


"""
@name:
	aBam2SCNV
@description:
	Call CNV with paired control samples
@procs:
	`pBamDir`       : The directory with the bam files
	`pSampleInfo`   : The sample information
	`pCNVkitPrepare`: Prepare the region/coverage file for CNV calling
	`pCNVkitCov`    : Get the coverage file for each bam file.
	`pCNNDir`       : Put all coverage file into a directory.
	`pCNVkitRef`    : Aggregate the reference cnn file from controls.
	`pCNVkitFix`    : Fix the coverage for each sample using the reference cnn file.
	`pCNVkitSeg`    : Segment the regions
	`pCNVkitCall`   : Call CNVs.
	`pCNVkitScatter`: Draw scatter plot for CNVs.
	`pCNVkitDiagram`: Draw diagram plot for CNVs.
	`pCNVkitHeatmap`: Draw heatmap for CNVs.
	`pCNVkitReport` : Generate a report for the results.
	`pCNVkit2Vcf`   : Export the results to VCF file.
@modules
	`plots`: Draw plots
		- ... -> `pCNVkitFix, pCNVkitSeg` -> `pCNVkitScatter[!], pCNVkitDiagram[!], pCNVkitHeatmap[!], pCNVkitReport[!]`
"""
aBam2SCNV = ProcSet(
	pFile2Proc.copy(id = 'pBamDir'),     # single job
	pFile2Proc.copy(id = 'pSampleInfo'), # single job
	pCNVkitPrepare,   # prepare files
	pCNVkitCov,       # each sample       , in: bam file (target required), out: cnn file
	pFiles2Dir.copy(id = 'pCNNDir'),  # put all cnn files in a directory
	pCNVkitRef,       # all normal samples, in: all normal cnn files      , out: refcnn
	pCNVkitFix,       # each tumor sample , in: tumor cnn, refcnn         , out: cnr
	pCNVkitSeg,       # each tumor sample , in: cnr                       , out: cns
	pCNVkitCall,      # each tumor sample , in: cns                       , out: callcns
	pCNVkitScatter,   # each tumor sample , in: cnr, cns                  , out: plots
	pCNVkitDiagram,   # each tumor sample , in: cnr, cns                  , out: plots
	pCNVkitHeatmap,   # all tumor samples , in: cnfiles                   , out: plots
	pCNVkitReport,    # each tumor sample , in: cnr, cns                  , out: reports
	pCNVkit2Vcf,      # each tumor sample , in: callcns                   , out: vcf,
	depends = False
)
# defaults
aBam2SCNV.pBamDir.runner     = 'local'
aBam2SCNV.pSampleInfo.runner = 'local'
# delegates
aBam2SCNV.delegate('args.ref', 'pCNVkitPrepare, pCNVkitRef')
aBam2SCNV.delegate('args.cnvkit', 'pCNVkit*')
aBam2SCNV.delegate('args.nthread', 'pCNVkitCov, pCNVkitSeg')
# depends
aBam2SCNV.starts                 = 'pBamDir, pSampleInfo'
aBam2SCNV.ends                   = 'pCNVkit2Vcf'
aBam2SCNV.pCNVkitPrepare.depends = aBam2SCNV.starts
aBam2SCNV.pCNVkitCov.depends     = aBam2SCNV['pBamDir, pSampleInfo, pCNVkitPrepare']
aBam2SCNV.pCNNDir.depends        = aBam2SCNV.pCNVkitCov
aBam2SCNV.pCNVkitRef.depends     = aBam2SCNV['pCNNDir, pSampleInfo']
aBam2SCNV.pCNVkitFix.depends     = aBam2SCNV['pCNNDir, pSampleInfo, pCNVkitRef']
aBam2SCNV.pCNVkitSeg.depends     = aBam2SCNV.pCNVkitFix
aBam2SCNV.pCNVkitCall.depends    = aBam2SCNV.pCNVkitSeg
aBam2SCNV.pCNVkit2Vcf.depends    = aBam2SCNV.pCNVkitCall
# input
aBam2SCNV.pCNVkitPrepare.input = lambda ch_bamdir, ch_saminfo: [
	Channel.create(
		SampleInfo(ch_saminfo.get()).toChannel(ch_bamdir.get())
	).unique().flatten()]
aBam2SCNV.pCNVkitCov.input     = lambda ch_bamdir, ch_saminfo, ch_target: \
	Channel.create(
		SampleInfo(ch_saminfo.get()).toChannel(ch_bamdir.get())
	).unique().cbind(ch_target)
aBam2SCNV.pCNNDir.input        = lambda ch: [ch.flatten()]
aBam2SCNV.pCNVkitRef.input     = lambda ch_covs, ch_saminfo: [
	Channel.create(
		SampleInfo(ch_saminfo.get()).toChannel(ch_covs.get(), paired = True, raiseExc = False)
	).colAt(1).unique().map(
		lambda x: (x[0].rpartition('.')[0] + '.target.cnn', x[0].rpartition('.')[0] + '.antitarget.cnn')
	).flatten()]
aBam2SCNV.pCNVkitFix.input     = lambda ch_covs, ch_saminfo, ch_ref: \
	Channel.create(
		SampleInfo(ch_saminfo.get()).toChannel(ch_covs.get(), paired = True)
	).colAt(0).unique().map(
		lambda x: (x[0].rpartition('.')[0] + '.target.cnn', x[0].rpartition('.')[0] + '.antitarget.cnn')
	).cbind(ch_ref)
aBam2SCNV.pCNVkitHeatmap.input = lambda ch: [ch.flatten()]
# module
@aBam2SCNV.module
def aBam2SCNV_plots(ps):
	ps.ends = 'pCNVkitScatter, pCNVkitDiagram, pCNVkitHeatmap, pCNVkitReport pCNVkit2Vcf'

	ps['pCNVkitScatter'].depends = ps['pCNVkitFix, pCNVkitSeg']
	ps['pCNVkitDiagram'].depends = ps['pCNVkitFix, pCNVkitSeg']
	ps['pCNVkitHeatmap'].depends = ps['pCNVkitSeg']
	ps['pCNVkitReport'].depends  = ps['pCNVkitFix, pCNVkitSeg']


"""
@name:
	aBam2GCNV
@description:
	Call CNV with a flat reference coverage file.
@procs:
	`pBamDir`       : The directory with the bam files
	`pSampleInfo`   : The sample information
	`pCNVkitPrepare`: Prepare the region/coverage file for CNV calling
	`pCNVkitCov`    : Get the coverage file for each bam file.
	`pCNNDir`       : Put all coverage file into a directory.
	`pCNVkitFlatRef`: Generate a flat reference coverage file.
	`pCNVkitFix`    : Fix the coverage for each sample using the reference cnn file.
	`pCNVkitSeg`    : Segment the regions
	`pCNVkitCall`   : Call CNVs.
	`pCNVkitScatter`: Draw scatter plot for CNVs.
	`pCNVkitDiagram`: Draw diagram plot for CNVs.
	`pCNVkitHeatmap`: Draw heatmap for CNVs.
	`pCNVkitReport` : Generate a report for the results.
	`pCNVkit2Vcf`   : Export the results to VCF file.
@modules
	`plots`: Draw plots
		- ... -> `pCNVkitFix, pCNVkitSeg` -> `pCNVkitScatter[!], pCNVkitDiagram[!], pCNVkitHeatmap[!], pCNVkitReport[!]`
"""
aBam2GCNV = ProcSet(
	pFile2Proc.copy(id = 'pBamDir'),     # single job g
	pFile2Proc.copy(id = 'pSampleInfo'), # single job g
	pCNVkitPrepare,    # prepare files
	pCNVkitCov,       # each sample       , in: bam file (target required), out: cnn file
	pFiles2Dir.copy(id = 'pCNNDir'),  # put all cnn files in a directory
	pCNVkitFlatRef,   # flat reference    , in: target file               , out: refcnn
	pCNVkitFix,       # each tumor sample , in: tumor cnn, refcnn         , out: cnr
	pCNVkitSeg,       # each tumor sample , in: cnr                       , out: cns
	pCNVkitCall,      # each tumor sample , in: cns                       , out: callcns
	pCNVkitScatter,   # each tumor sample , in: cnr, cns                  , out: plots
	pCNVkitDiagram,   # each tumor sample , in: cnr, cns                  , out: plots
	pCNVkitHeatmap,   # all tumor samples , in: cnfiles                   , out: plots
	pCNVkitReport,    # each tumor sample , in: cnr, cns                  , out: reports
	pCNVkit2Vcf,      # each tumor sample , in: callcns                   , out: vcf,
	depends = False
)
# defaults
aBam2GCNV.pBamDir.runner     = 'local'
aBam2GCNV.pSampleInfo.runner = 'local'
# delegates
aBam2GCNV.delegate('args.ref', 'pCNVkitPrepare, pCNVkitFlatRef')
aBam2GCNV.delegate('args.nthread', 'pCNVkitCov, pCNVkitSeg')
aBam2GCNV.delegate('args.cnvkit', 'pCNVkit*')
# depends
aBam2GCNV.starts                 = 'pBamDir, pSampleInfo'
aBam2GCNV.ends                   = 'pCNVkit2Vcf'
aBam2GCNV.pCNVkitPrepare.depends = aBam2GCNV.starts
aBam2GCNV.pCNVkitCov.depends     = aBam2GCNV['pBamDir, pSampleInfo, pCNVkitPrepare']
aBam2GCNV.pCNNDir.depends        = aBam2GCNV.pCNVkitCov
aBam2GCNV.pCNVkitFlatRef.depends = aBam2GCNV.pCNVkitPrepare
aBam2GCNV.pCNVkitFix.depends     = aBam2GCNV['pCNNDir, pSampleInfo, pCNVkitFlatRef']
aBam2GCNV.pCNVkitSeg.depends     = aBam2GCNV.pCNVkitFix
aBam2GCNV.pCNVkitCall.depends    = aBam2GCNV.pCNVkitSeg
aBam2GCNV.pCNVkit2Vcf.depends    = aBam2GCNV.pCNVkitCall
# input
#aBam2GCNV.pCNVkitAccess.input  = lambda ch: path.basename(ch.get()).split('.')[:1]
aBam2GCNV.pCNVkitPrepare.input = lambda ch_bamdir, ch_saminfo: [
	Channel.create(
		SampleInfo(ch_saminfo.get()).toChannel(ch_bamdir.get())
	).unique().flatten()]
aBam2GCNV.pCNVkitCov.input     = lambda ch_bamdir, ch_saminfo, ch_target: \
	Channel.create(
		SampleInfo(ch_saminfo.get()).toChannel(ch_bamdir.get())
	).unique().cbind(ch_target)
aBam2GCNV.pCNNDir.input        = lambda ch: [ch.flatten()]
aBam2GCNV.pCNVkitFix.input     = lambda ch_covs, ch_saminfo, ch_ref: \
	Channel.create(
		SampleInfo(ch_saminfo.get()).toChannel(ch_covs.get())
	).colAt(0).unique().map(
		lambda x: (x[0].rpartition('.')[0] + '.target.cnn', x[0].rpartition('.')[0] + '.antitarget.cnn')
	).cbind(ch_ref)
aBam2GCNV.pCNVkitHeatmap.input = lambda ch: [ch.flatten()]
# module
@aBam2GCNV.module
def aBam2GCNV_plots(ps):
	ps.ends = 'pCNVkitScatter, pCNVkitDiagram, pCNVkitHeatmap, pCNVkitReport'
	ps['pCNVkitScatter'].depends = ps['pCNVkitFix, pCNVkitSeg']
	ps['pCNVkitDiagram'].depends = ps['pCNVkitFix, pCNVkitSeg']
	ps['pCNVkitHeatmap'].depends = ps['pCNVkitSeg']
	ps['pCNVkitReport'].depends  = ps['pCNVkitFix, pCNVkitSeg']

"""
@name:
	aVcfs2Maf
@description:
	Convert vcf files to maf files and merge them.
@procs:
	`pVcf2Maf`  : Convert single vcf file to maf file
	`pFiles2Dir`: Put the maf files into a directory
	`pMafMerge` : Merge the maf files.
"""
aVcfs2Maf = ProcSet(
	pVcf2Maf,
	pFiles2Dir,
	pMafMerge
)
aVcfs2Maf.pFiles2Dir.input = lambda ch: [ch.flatten()]



"""
./topiary \
  --vcf somatic.vcf \
  --mhc-predictor netmhcpan \
  --mhc-alleles HLA-A*02:01,HLA-B*07:02 \
  --ic50-cutoff 500 \
  --percentile-cutoff 2.0 \
  --mhc-epitope-lengths 8-11 \
  --rna-gene-fpkm-tracking-file genes.fpkm_tracking \
  --rna-min-gene-expression 4.0 \
  --rna-transcript-fpkm-tracking-file isoforms.fpkm_tracking \
  --rna-min-transcript-expression 1.5 \
  --output-csv epitopes.csv
"""
import re
from os import environ
from pathlib import Path
from pyppl import Box
from bioprocs.utils import shell2 as shell
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile        = {{i.infile | quote}}
mhcallele     = {{i.mhcallele | quote}}
gexpr         = {{i.gexpr | quote}}
texpr         = {{i.texpr | quote}}
outfile       = Path({{o.outfile | quote}})
neatfile      = outfile.with_suffix('.neat.txt')
topiary       = {{args.topiary | quote}}
netmhc        = {{args.netmhc | quote}}
netmhcpan     = {{args.netmhcpan | quote}}
netmhciipan   = {{args.netmhciipan | quote}}
netmhccons    = {{args.netmhccons | quote}}
smm           = {{args.smm | quote}}
smm_pmbec     = {{args.smm_pmbec | quote}}
mhc_predictor = {{args.mhc_predictor | quote}}
genome        = {{args.genome | quote}}
params        = {{args.params | repr}}
wildtype      = {{args.wildtype | repr}}
tmpdir        = Path({{args.tmpdir | quote}}) / '.'.join([
	{{proc.id | quote}}, {{proc.tag | quote}}, {{proc.suffix | quote}}, {{job.index | quote}}])
tmpdir.mkdir(exist_ok = True, parents = True)

# check if we have downloaded annotation data for the genome
gmaps = {'hg19': 'GRCh37', 'hg38': 'GRCh38'}
datadir = Path.home().joinpath('.cache', 'pyensembl')
if not datadir.joinpath(genome).is_dir() and not datadir.joinpath(gmaps.get(genome, genome)).is_dir():
	raise RuntimeError("You don't have annotation data for genome {}{} installed. "
		"Either you run 'pyensembl install' first or "
		"specify 'params.download_reference_genome_data = True'. "
		"If you have it installed somewhere else, make a symbolic link to {}".format(genome, ('/' + gmaps[genome]) if genome in gmaps else '', datadir))
if not datadir.joinpath(genome).is_dir() and datadir.joinpath(gmaps.get(genome, genome)).is_dir():
	genome = gmaps[genome]
datadir = datadir / genome

shell.load_config(topiary = topiary)
if infile.endswith('.vcf') or infile.endswith('.vcf.gz'):
	params.vcf = infile
else:
	params.maf = infile

if Path(mhcallele).is_file():
	mhcallelefile = outfile.parent.joinpath('mhcalleles.txt')
	with open(mhcallele, 'r') as fin, open(mhcallelefile, 'w') as fout:
		for line in fin:
			fout.write(line.replace('*', ''))
	params['mhc-alleles-file'] = mhcallelefile
else:
	params['mhc-alleles'] = mhcallele.replace('*', '')

if gexpr:
	params['rna-gene-fpkm-tracking-file'] = gexpr
if texpr:
	params['rna-transcript-fpkm-tracking-file'] = texpr

params.genome = genome
params['output-csv'] = outfile
params['mhc-predictor'] = mhc_predictor

# make sure those mhc-predictors are in PATH
PATHs = set()
for mhcpred in (netmhc, netmhcpan, netmhciipan, netmhccons, smm, smm_pmbec):
	if '/' in mhcpred:
		PATHs.add(str(Path(mhcpred).parent))

params._env = Box(PATH = environ['PATH'] + ':' + ':'.join(PATHs))

shell.fg.topiary(**params)

if not wildtype:
	reader = TsvReader(outfile, comment = '###')
	writer = TsvWriter(neatfile)
	writer.cnames = ['HLA_allele', 'mt_peptide', 'mt_affinity', 'gene']
	writer.writeHead()
	for r in reader:
		writer.write([r.allele, r.peptide, r.affinity, r.gene])
else:
	if mhc_predictor not in ('netmhc', 'netmhcpan', 'netmhciipan', 'netmhccons', 'smm', 'smm_pmbec'):
		raise ValueError("Calculations for wildtype peptides are only available with local predictors.")

	tmpfile = outfile.with_suffix('.topiary.txt')
	shell.mv(outfile, tmpfile)

	wildpeps = set()
	mutpeps = {}
	tpreader = TsvReader(tmpfile, comment = '###')
	for r in tpreader:
		if r.effect_type != 'Substitution':
			# I don't know how to get the wildtype peptides if it is not a substitution
			continue
		# parse effect: p.N84S
		m = re.match(r'^p\.([A-Z])\d+([A-Z])$', r.effect)
		if not m:
			continue
		wildpep = r.peptide
		index = int(r.mutation_start_in_peptide)
		if wildpep[index] != m.group(2):
			continue
		wildpep = wildpep[:index] + m.group(1) + wildpep[(index+1):]
		mutpeps[r.peptide + '\t' + r.allele] = wildpep
		wildpeps.add(wildpep)

	def run_netmhc():
		pass

	def run_netmhcpan():
		shell.load_config(netmhcpan = netmhcpan)
		mhcallele2 = params['mhc-alleles'] if 'mhc-alleles' in params else ','.join(
			allele for allele in Path(params['mhc-alleles-file']).read_text().splitlines() if allele
		)

		wildfile = outfile.parent / 'wildtype.peptides.txt'
		wildfile.write_text('\n'.join(wildpeps))
		xlsfile = outfile.parent / 'wildtype.binding.txt'
		nparams = Box(
			a = mhcallele2, v = True, BA = True, inptype = 1, f = wildfile, _prefix = '-',
			xls = True, xlsfile = xlsfile)
		shell.fg.netmhcpan(**nparams)

		if not xlsfile.is_file():
			raise RuntimeError("Failed to run netmhcpan, output file not generated.")
		# read the output
		"""
					HLA-A24:02					HLA-A29:02
		Pos	Peptide	ID	core	icore	1-log50k	nM	Rank	core	icore	1-log50k	nM	Rank
		0	LYLPALWFH	PEPLIST	LYLPALWFH	LYLPALWFH	0.1353	11560.5488	6.1138	LYLPALWFH	LYLPALWFH	0.4137	568.6087	1.1231
		0	RRQRRQRRW	PEPLIST	RRQRRQRRW	RRQRRQRRW	0.0788	21311.8301	12.3392	RRQRRQRRW	RRQRRQRRW	0.0308	35829.9805	47.6206
		"""
		with xlsfile.open('r') as f:
			alleles = [allele.replace('HLA-A', 'HLA-A*').replace('HLA-B', 'HLA-B*').replace('HLA-C', 'HLA-C*')
				for allele in f.readline().strip().split('\t') if allele]
		reader = TsvReader(xlsfile, comment = '\t\t\t')
		wildbindings = {}
		for r in reader:
			peptide = r[1]
			for i, hla in enumerate(alleles):
				wildbindings[peptide + '\t' + hla] = float(r[7 + i*5])

		writer = TsvWriter(outfile)
		writer.cnames = tpreader.cnames + ['wildpeptide', 'wildaffinity', 'deltaaffinity']
		writer.writeHead()
		nwriter = TsvWriter(neatfile)
		nwriter.cnames = ['HLA_allele', 'mt_peptide', 'mt_affinity', 'wt_peptide', 'wt_affinity', 'delta_affinity', 'gene']
		nwriter.writeHead()
		tpreader.rewind()
		for r in tpreader:
			r.wildpeptide = mutpeps.get(r.peptide + '\t' + r.allele, '-')
			r.wildaffinity = wildbindings.get(r.wildpeptide + '\t' + r.allele, '-')
			if r.wildaffinity != '-':
				r.deltaaffinity = float(r.affinity) - r.wildaffinity
			else:
				r.deltaaffinity = '-'
			nwriter.write([	r.allele, r.peptide, r.affinity, r.wildpeptide,
							r.wildaffinity, r.deltaaffinity, r.gene])
			writer.write(r)

	def run_netmhciipan():
		pass

	def run_netmhccons():
		pass

	def run_smm():
		pass

	def run_smm_pmbec():
		pass

	runner = {
		'netmhc'      : run_netmhc,
		'netmhcpan'   : run_netmhcpan,
		'netmhciipan' : run_netmhciipan,
		'netmhccons'  : run_netmhccons,
		'smm'         : run_smm,
		'smm-pmbec'   : run_smm_pmbec,
	}
	runner.get(mhc_predictor)()



from os import path
from pyppl import Box
from bioprocs.utils import regionOverlap, log2pyppl
from bioprocs.utils.tsvio2 import TsvJoin

infiles = {{i.infiles}}
inopts  = {{args.inopts | repr}}
outopts = {{args.outopts | repr}}
debug   = {{args.debug}}
outfile = {{o.outfile | quote}}
sefile  = {{job.errfile | quote}}
helper  = {{args.helper | repr}}
if not isinstance(helper, list):
	helper = [helper]

compare = TsvJoin.compare
helper  = [line for line in helper if line]
exec('\n'.join(helper), globals())

do      = {{args.do}}
match   = {{args.match}}

tj = TsvJoin(*infiles, **inopts)
tj.debug = debug
outopts['headCallback'] = {{args.outopts.get('headCallback', True)}}
tj.join(do, outfile, match, outopts)

with open(sefile, 'r') as f:
	f.flush()
	i = 0
	for line in f:
		if line.startswith('pyppl.log'):
			continue
		log2pyppl(line.rstrip(), 'warning')
		i += 1
		if i >= 20:
			break
	if i > 0:
		log2pyppl('See more in {}'.format(sefile), 'warning')
